from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.comment import Comment
from ..arctrl_core.data import Data
from ..arctrl_core.data_file import DataFile
from ..arctrl_core.uri import URIModule_toString
from ..fable_library.list import (choose, of_array, FSharpList, singleton)
from ..fable_library.option import map
from ..fable_library.seq import map as map_1
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IOptionalGetter, string, resize_array, IGetters)
from ..thoth_json_core.encode import list_1 as list_1_1
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_1, decoder as decoder_1, ROCrate_encoder as ROCrate_encoder_2, ROCrate_decoder as ROCrate_decoder_2, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_2)
from .context.rocrate.isa_data_context import context_jsonvalue
from .data_file import (ISAJson_encoder as ISAJson_encoder_1, ISAJson_decoder as ISAJson_decoder_1, ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_1)
from .decode import (Decode_uri, Decode_objectNoAdditionalProperties)
from .encode import (try_include, try_include_seq)
from .idtable import encode
from .string_table import (encode_string, decode_string)

__A_ = TypeVar("__A_")

def encoder(d: Data) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], d: Any=d) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2164(value: str, d: Any=d) -> IEncodable:
        class ObjectExpr2163(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2163()

    def _arrow2166(value_2: str, d: Any=d) -> IEncodable:
        class ObjectExpr2165(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2165()

    def _arrow2168(value_4: str, d: Any=d) -> IEncodable:
        class ObjectExpr2167(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2167()

    def _arrow2170(value_6: str, d: Any=d) -> IEncodable:
        class ObjectExpr2169(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_6)

        return ObjectExpr2169()

    def _arrow2171(comment: Comment, d: Any=d) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2164, d.ID), try_include("name", _arrow2166, d.Name), try_include("dataType", ISAJson_encoder_1, d.DataType), try_include("format", _arrow2168, d.Format), try_include("selectorFormat", _arrow2170, d.SelectorFormat), try_include_seq("comments", _arrow2171, d.Comments)]))
    class ObjectExpr2172(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], d: Any=d) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_4))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_4.encode_object(arg)

    return ObjectExpr2172()


def _arrow2179(get: IGetters) -> Data:
    def _arrow2173(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2174(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2175(__unit: None=None) -> DataFile | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("dataType", ISAJson_decoder_1)

    def _arrow2176(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("format", string)

    def _arrow2177(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("selectorFormat", Decode_uri)

    def _arrow2178(__unit: None=None) -> Array[Comment] | None:
        arg_11: Decoder_1[Array[Comment]] = resize_array(decoder_1)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("comments", arg_11)

    return Data(_arrow2173(), _arrow2174(), _arrow2175(), _arrow2176(), _arrow2177(), _arrow2178())


decoder: Decoder_1[Data] = object(_arrow2179)

def compressed_encoder(string_table: Any, d: Data) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, d: Any=d) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2181(s: str, string_table: Any=string_table, d: Any=d) -> IEncodable:
        return encode_string(string_table, s)

    def _arrow2182(s_1: str, string_table: Any=string_table, d: Any=d) -> IEncodable:
        return encode_string(string_table, s_1)

    def _arrow2183(s_2: str, string_table: Any=string_table, d: Any=d) -> IEncodable:
        return encode_string(string_table, s_2)

    def _arrow2184(s_3: str, string_table: Any=string_table, d: Any=d) -> IEncodable:
        return encode_string(string_table, s_3)

    def _arrow2185(comment: Comment, string_table: Any=string_table, d: Any=d) -> IEncodable:
        return encoder_1(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("i", _arrow2181, d.ID), try_include("n", _arrow2182, d.Name), try_include("d", ISAJson_encoder_1, d.DataType), try_include("f", _arrow2183, d.Format), try_include("s", _arrow2184, d.SelectorFormat), try_include_seq("c", _arrow2185, d.Comments)]))
    class ObjectExpr2186(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], string_table: Any=string_table, d: Any=d) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers.encode_object(arg)

    return ObjectExpr2186()


def compressed_decoder(string_table: Array[str]) -> Decoder_1[Data]:
    def _arrow2193(get: IGetters, string_table: Any=string_table) -> Data:
        def _arrow2187(__unit: None=None) -> str | None:
            arg_1: Decoder_1[str] = decode_string(string_table)
            object_arg: IOptionalGetter = get.Optional
            return object_arg.Field("i", arg_1)

        def _arrow2188(__unit: None=None) -> str | None:
            arg_3: Decoder_1[str] = decode_string(string_table)
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("n", arg_3)

        def _arrow2189(__unit: None=None) -> DataFile | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("d", ISAJson_decoder_1)

        def _arrow2190(__unit: None=None) -> str | None:
            arg_7: Decoder_1[str] = decode_string(string_table)
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("f", arg_7)

        def _arrow2191(__unit: None=None) -> str | None:
            arg_9: Decoder_1[str] = decode_string(string_table)
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("s", arg_9)

        def _arrow2192(__unit: None=None) -> Array[Comment] | None:
            arg_11: Decoder_1[Array[Comment]] = resize_array(decoder_1)
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("c", arg_11)

        return Data(_arrow2187(), _arrow2188(), _arrow2189(), _arrow2190(), _arrow2191(), _arrow2192())

    return object(_arrow2193)


def ROCrate_genID(d: Data) -> str:
    match_value: str | None = d.ID
    if match_value is None:
        match_value_1: str | None = d.Name
        if match_value_1 is None:
            return "#EmptyData"

        else: 
            return replace(match_value_1, " ", "_")


    else: 
        return URIModule_toString(match_value)



def ROCrate_encoder(oa: Data) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], oa: Any=oa) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2197(__unit: None=None, oa: Any=oa) -> IEncodable:
        value: str = ROCrate_genID(oa)
        class ObjectExpr2196(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2196()

    class ObjectExpr2198(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            return helpers_1.encode_string("Data")

    def _arrow2200(value_2: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2199(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_2)

        return ObjectExpr2199()

    def _arrow2201(value_4: DataFile, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_1(value_4)

    def _arrow2203(value_5: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2202(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_5)

        return ObjectExpr2202()

    def _arrow2205(value_7: str, oa: Any=oa) -> IEncodable:
        class ObjectExpr2204(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
                return helpers_4.encode_string(value_7)

        return ObjectExpr2204()

    def _arrow2206(comment: Comment, oa: Any=oa) -> IEncodable:
        return ROCrate_encoder_2(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2197()), ("@type", list_1_1(singleton(ObjectExpr2198()))), try_include("name", _arrow2200, oa.Name), try_include("type", _arrow2201, oa.DataType), try_include("encodingFormat", _arrow2203, oa.Format), try_include("usageInfo", _arrow2205, oa.SelectorFormat), try_include_seq("comments", _arrow2206, oa.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2207(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any], oa: Any=oa) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr2207()


def _arrow2214(get: IGetters) -> Data:
    def _arrow2208(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2209(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2210(__unit: None=None) -> DataFile | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("type", ROCrate_decoder_1)

    def _arrow2211(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("encodingFormat", string)

    def _arrow2212(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("usageInfo", Decode_uri)

    def _arrow2213(__unit: None=None) -> Array[Comment] | None:
        arg_11: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoder_2)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("comments", arg_11)

    return Data(_arrow2208(), _arrow2209(), _arrow2210(), _arrow2211(), _arrow2212(), _arrow2213())


ROCrate_decoder: Decoder_1[Data] = object(_arrow2214)

def ISAJson_encoder(id_map: Any | None, oa: Data) -> IEncodable:
    def f(oa_1: Data, id_map: Any=id_map, oa: Any=oa) -> IEncodable:
        def chooser(tupled_arg: tuple[str, IEncodable | None], oa_1: Any=oa_1) -> tuple[str, IEncodable] | None:
            def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping, tupled_arg[1])

        def _arrow2218(value: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2217(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr2217()

        def _arrow2220(value_2: str, oa_1: Any=oa_1) -> IEncodable:
            class ObjectExpr2219(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_2)

            return ObjectExpr2219()

        def _arrow2221(comment: Comment, oa_1: Any=oa_1) -> IEncodable:
            return ISAJson_encoder_2(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([try_include("@id", _arrow2218, ROCrate_genID(oa_1)), try_include("name", _arrow2220, oa_1.Name), try_include("type", ISAJson_encoder_1, oa_1.DataType), try_include_seq("comments", _arrow2221, oa_1.Comments)]))
        class ObjectExpr2222(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], oa_1: Any=oa_1) -> Any:
                def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_2))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
                return helpers_2.encode_object(arg)

        return ObjectExpr2222()

    if id_map is not None:
        def _arrow2223(d_1: Data, id_map: Any=id_map, oa: Any=oa) -> str:
            return ROCrate_genID(d_1)

        return encode(_arrow2223, f, oa, id_map)

    else: 
        return f(oa)



ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "name", "type", "comments", "@type", "@context"])

def _arrow2228(get: IGetters) -> Data:
    def _arrow2224(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("@id", Decode_uri)

    def _arrow2225(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("name", string)

    def _arrow2226(__unit: None=None) -> DataFile | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("type", ISAJson_decoder_1)

    def _arrow2227(__unit: None=None) -> Array[Comment] | None:
        arg_7: Decoder_1[Array[Comment]] = resize_array(ISAJson_decoder_2)
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("comments", arg_7)

    return Data(_arrow2224(), _arrow2225(), _arrow2226(), None, None, _arrow2227())


ISAJson_decoder: Decoder_1[Data] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow2228)

__all__ = ["encoder", "decoder", "compressed_encoder", "compressed_decoder", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ....arctrl_core.Process.process_output import ProcessOutput
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.Process.process_output import (ISAJson_decoder, ISAJson_encoder, ROCrate_decoder, ROCrate_encoder)
from ....fable_library.option import default_arg
from ....fable_library.result import FSharpResult_2
from ....fable_library.string_ import (to_text, printf)
from ....thoth_json_core.types import IEncodable
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string

def ARCtrl_Process_ProcessOutput__ProcessOutput_fromISAJsonString_Static_Z721C83C5(s: str) -> ProcessOutput:
    match_value: FSharpResult_2[ProcessOutput, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Process_ProcessOutput__ProcessOutput_toISAJsonString_Static_Z3B036AA(spaces: int | None=None, use_idreferencing: bool | None=None) -> Callable[[ProcessOutput], str]:
    id_map: Any | None = dict([]) if default_arg(use_idreferencing, False) else None
    def _arrow3824(f: ProcessOutput, spaces: Any=spaces, use_idreferencing: Any=use_idreferencing) -> str:
        value: IEncodable = ISAJson_encoder(id_map, f)
        return to_string(default_spaces(spaces), value)

    return _arrow3824


def ARCtrl_Process_ProcessOutput__ProcessOutput_toISAJsonString_Z3B036AA(this: ProcessOutput, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
    return ARCtrl_Process_ProcessOutput__ProcessOutput_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(this)


def ARCtrl_Process_ProcessOutput__ProcessOutput_fromROCrateJsonString_Static_Z721C83C5(s: str) -> ProcessOutput:
    match_value: FSharpResult_2[ProcessOutput, str] = Decode_fromString(ROCrate_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Process_ProcessOutput__ProcessOutput_toROCrateJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[ProcessOutput], str]:
    def _arrow3825(f: ProcessOutput, spaces: Any=spaces) -> str:
        value: IEncodable = ROCrate_encoder(f)
        return to_string(default_spaces(spaces), value)

    return _arrow3825


def ARCtrl_Process_ProcessOutput__ProcessOutput_toROCrateJsonString_71136F3F(this: ProcessOutput, spaces: int | None=None) -> str:
    return ARCtrl_Process_ProcessOutput__ProcessOutput_toROCrateJsonString_Static_71136F3F(spaces)(this)


__all__ = ["ARCtrl_Process_ProcessOutput__ProcessOutput_fromISAJsonString_Static_Z721C83C5", "ARCtrl_Process_ProcessOutput__ProcessOutput_toISAJsonString_Static_Z3B036AA", "ARCtrl_Process_ProcessOutput__ProcessOutput_toISAJsonString_Z3B036AA", "ARCtrl_Process_ProcessOutput__ProcessOutput_fromROCrateJsonString_Static_Z721C83C5", "ARCtrl_Process_ProcessOutput__ProcessOutput_toROCrateJsonString_Static_71136F3F", "ARCtrl_Process_ProcessOutput__ProcessOutput_toROCrateJsonString_71136F3F"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1850() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.ROCrate.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("CreativeWork", string_type), ("about", string_type), ("conforms_to", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    CreativeWork: str
    about: str
    conforms_to: str

IContext_reflection = _expr1850

def _arrow1853(__unit: None=None) -> IEncodable:
    class ObjectExpr1851(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("https://w3id.org/ro/crate/1.1")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("@id", ObjectExpr1851())])
    class ObjectExpr1852(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_1))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_1.encode_object(arg)

    return ObjectExpr1852()


conforms_to_jsonvalue: IEncodable = _arrow1853()

def _arrow1861(__unit: None=None) -> IEncodable:
    class ObjectExpr1854(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1855(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("http://purl.org/nfdi4plants/ontology/")

    class ObjectExpr1856(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:CreativeWork")

    class ObjectExpr1857(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:about")

    class ObjectExpr1858(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("http://purl.org/dc/terms/conformsTo")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1854()), ("arc", ObjectExpr1855()), ("CreativeWork", ObjectExpr1856()), ("about", ObjectExpr1857()), ("conformsTo", ObjectExpr1858())])
    class ObjectExpr1859(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_5))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_5.encode_object(arg)

    return ObjectExpr1859()


context_jsonvalue: IEncodable = _arrow1861()

__all__ = ["IContext_reflection", "conforms_to_jsonvalue", "context_jsonvalue"]




from __future__ import annotations
from typing import Any
from ...thoth_json_core.types import IDecoderHelpers_1
from ...thoth_json_python.decode import Decode_helpers

helpers: IDecoderHelpers_1[Any] = Decode_helpers

__all__ = ["helpers"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.publication import Publication
from ...arctrl_json.encode import default_spaces
from ...arctrl_json.publication import (decoder as decoder_1, encoder, ROCrate_decoder, ROCrate_encoder, ISAJson_decoder, ISAJson_encoder)
from ...fable_library.option import default_arg
from ...fable_library.result import FSharpResult_2
from ...fable_library.string_ import (to_text, printf)
from ...thoth_json_core.types import IEncodable
from ...thoth_json_python.decode import Decode_fromString
from ...thoth_json_python.encode import to_string

def ARCtrl_Publication__Publication_fromJsonString_Static_Z721C83C5(s: str) -> Publication:
    match_value: FSharpResult_2[Publication, str] = Decode_fromString(decoder_1, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Publication__Publication_toJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[Publication], str]:
    def _arrow3800(obj: Publication, spaces: Any=spaces) -> str:
        value: IEncodable = encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3800


def ARCtrl_Publication__Publication_ToJsonString_71136F3F(this: Publication, spaces: int | None=None) -> str:
    return ARCtrl_Publication__Publication_toJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_Publication__Publication_fromROCrateJsonString_Static_Z721C83C5(s: str) -> Publication:
    match_value: FSharpResult_2[Publication, str] = Decode_fromString(ROCrate_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Publication__Publication_toROCrateJsonString_Static_71136F3F(spaces: int | None=None) -> Callable[[Publication], str]:
    def _arrow3801(obj: Publication, spaces: Any=spaces) -> str:
        value: IEncodable = ROCrate_encoder(obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3801


def ARCtrl_Publication__Publication_ToROCrateJsonString_71136F3F(this: Publication, spaces: int | None=None) -> str:
    return ARCtrl_Publication__Publication_toROCrateJsonString_Static_71136F3F(spaces)(this)


def ARCtrl_Publication__Publication_fromISAJsonString_Static_Z721C83C5(s: str) -> Publication:
    match_value: FSharpResult_2[Publication, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Publication__Publication_toISAJsonString_Static_Z3B036AA(spaces: int | None=None, use_idreferencing: bool | None=None) -> Callable[[Publication], str]:
    id_map: Any | None = dict([]) if default_arg(use_idreferencing, False) else None
    def _arrow3802(obj: Publication, spaces: Any=spaces, use_idreferencing: Any=use_idreferencing) -> str:
        value: IEncodable = ISAJson_encoder(id_map, obj)
        return to_string(default_spaces(spaces), value)

    return _arrow3802


def ARCtrl_Publication__Publication_ToISAJsonString_Z3B036AA(this: Publication, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
    return ARCtrl_Publication__Publication_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(this)


__all__ = ["ARCtrl_Publication__Publication_fromJsonString_Static_Z721C83C5", "ARCtrl_Publication__Publication_toJsonString_Static_71136F3F", "ARCtrl_Publication__Publication_ToJsonString_71136F3F", "ARCtrl_Publication__Publication_fromROCrateJsonString_Static_Z721C83C5", "ARCtrl_Publication__Publication_toROCrateJsonString_Static_71136F3F", "ARCtrl_Publication__Publication_ToROCrateJsonString_71136F3F", "ARCtrl_Publication__Publication_fromISAJsonString_Static_Z721C83C5", "ARCtrl_Publication__Publication_toISAJsonString_Static_Z3B036AA", "ARCtrl_Publication__Publication_ToISAJsonString_Z3B036AA"]




from __future__ import annotations
from abc import abstractmethod
from typing import (Generic, Protocol, TypeVar)
from .events import AbstractEventLoop
from .futures import (Future_1, Awaitable_1)

_T = TypeVar("_T")

class Task_1(Future_1, Awaitable_1, Generic[_T]):
    pass

class Task(Future_1, Awaitable_1):
    pass

class Coroutine_1(Awaitable_1, Generic[_T]):
    @abstractmethod
    def close(self) -> None:
        ...

    @abstractmethod
    def send(self, __arg0: _T) -> None:
        ...

    @abstractmethod
    def throw(self, __arg0: Exception) -> None:
        ...


class IExports(Protocol):
    @abstractmethod
    def get_event_loop(self) -> AbstractEventLoop:
        ...

    @abstractmethod
    def get_running_loop(self) -> AbstractEventLoop:
        ...





from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ...arctrl_core.data import Data
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Table.composite_cell import CompositeCell
from ...fable_library.list import (singleton, of_array, FSharpList)
from ...fable_library.seq import map
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.types import Array
from ...fable_library.util import (to_enumerable, IEnumerable_1)
from ...thoth_json_core.decode import (object, IRequiredGetter, string, index, IGetters)
from ...thoth_json_core.encode import list_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..data import (encoder as encoder_1, decoder as decoder_5, compressed_encoder, compressed_decoder)
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)
from ..string_table import (encode_string, decode_string)
from .oatable import (encode_oa, decode_oa)

__A_ = TypeVar("__A_")

def encoder(cc: CompositeCell) -> IEncodable:
    def oa_to_json_string(oa: OntologyAnnotation, cc: Any=cc) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    class ObjectExpr2743(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], cc: Any=cc) -> Any:
            return helpers_1.encode_string(cc.fields[0])

    class ObjectExpr2744(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], cc: Any=cc) -> Any:
            return helpers.encode_string(cc.fields[0])

    pattern_input: tuple[str, FSharpList[IEncodable]] = (("Term", singleton(oa_to_json_string(cc.fields[0])))) if (cc.tag == 0) else ((("Unitized", of_array([ObjectExpr2743(), oa_to_json_string(cc.fields[1])]))) if (cc.tag == 2) else ((("Data", singleton(encoder_1(cc.fields[0])))) if (cc.tag == 3) else (("FreeText", singleton(ObjectExpr2744())))))
    class ObjectExpr2745(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], cc: Any=cc) -> Any:
            return helpers_2.encode_string(pattern_input[0])

    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("celltype", ObjectExpr2745()), ("values", list_1(pattern_input[1]))])
    class ObjectExpr2746(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], cc: Any=cc) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values_1)
            return helpers_3.encode_object(arg)

    return ObjectExpr2746()


def _arrow2752(get: IGetters) -> CompositeCell:
    match_value: str
    object_arg: IRequiredGetter = get.Required
    match_value = object_arg.Field("celltype", string)
    def _arrow2747(__unit: None=None) -> str:
        arg_3: Decoder_1[str] = index(0, string)
        object_arg_1: IRequiredGetter = get.Required
        return object_arg_1.Field("values", arg_3)

    def _arrow2748(__unit: None=None) -> OntologyAnnotation:
        arg_5: Decoder_1[OntologyAnnotation] = index(0, OntologyAnnotation_decoder)
        object_arg_2: IRequiredGetter = get.Required
        return object_arg_2.Field("values", arg_5)

    def _arrow2749(__unit: None=None) -> str:
        arg_7: Decoder_1[str] = index(0, string)
        object_arg_3: IRequiredGetter = get.Required
        return object_arg_3.Field("values", arg_7)

    def _arrow2750(__unit: None=None) -> OntologyAnnotation:
        arg_9: Decoder_1[OntologyAnnotation] = index(1, OntologyAnnotation_decoder)
        object_arg_4: IRequiredGetter = get.Required
        return object_arg_4.Field("values", arg_9)

    def _arrow2751(__unit: None=None) -> Data:
        arg_11: Decoder_1[Data] = index(0, decoder_5)
        object_arg_5: IRequiredGetter = get.Required
        return object_arg_5.Field("values", arg_11)

    return CompositeCell(1, _arrow2747()) if (match_value == "FreeText") else (CompositeCell(0, _arrow2748()) if (match_value == "Term") else (CompositeCell(2, _arrow2749(), _arrow2750()) if (match_value == "Unitized") else (CompositeCell(3, _arrow2751()) if (match_value == "Data") else to_fail(printf("Error reading CompositeCell from json string: %A"))(match_value))))


decoder: Decoder_1[CompositeCell] = object(_arrow2752)

def encoder_compressed(string_table: Any, oa_table: Any, cc: CompositeCell) -> IEncodable:
    pattern_input: tuple[str, FSharpList[IEncodable]] = (("Term", singleton(encode_oa(oa_table, cc.fields[0])))) if (cc.tag == 0) else ((("Unitized", of_array([encode_string(string_table, cc.fields[0]), encode_oa(oa_table, cc.fields[1])]))) if (cc.tag == 2) else ((("Data", singleton(compressed_encoder(string_table, cc.fields[0])))) if (cc.tag == 3) else (("FreeText", singleton(encode_string(string_table, cc.fields[0]))))))
    values_1: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("t", encode_string(string_table, pattern_input[0])), ("v", list_1(pattern_input[1]))])
    class ObjectExpr2753(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cc: Any=cc) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values_1)
            return helpers.encode_object(arg)

    return ObjectExpr2753()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation]) -> Decoder_1[CompositeCell]:
    def _arrow2759(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table) -> CompositeCell:
        match_value: str
        arg_1: Decoder_1[str] = decode_string(string_table)
        object_arg: IRequiredGetter = get.Required
        match_value = object_arg.Field("t", arg_1)
        def _arrow2754(__unit: None=None) -> str:
            arg_3: Decoder_1[str] = index(0, decode_string(string_table))
            object_arg_1: IRequiredGetter = get.Required
            return object_arg_1.Field("v", arg_3)

        def _arrow2755(__unit: None=None) -> OntologyAnnotation:
            arg_5: Decoder_1[OntologyAnnotation] = index(0, decode_oa(oa_table))
            object_arg_2: IRequiredGetter = get.Required
            return object_arg_2.Field("v", arg_5)

        def _arrow2756(__unit: None=None) -> str:
            arg_7: Decoder_1[str] = index(0, decode_string(string_table))
            object_arg_3: IRequiredGetter = get.Required
            return object_arg_3.Field("v", arg_7)

        def _arrow2757(__unit: None=None) -> OntologyAnnotation:
            arg_9: Decoder_1[OntologyAnnotation] = index(1, decode_oa(oa_table))
            object_arg_4: IRequiredGetter = get.Required
            return object_arg_4.Field("v", arg_9)

        def _arrow2758(__unit: None=None) -> Data:
            arg_11: Decoder_1[Data] = index(0, compressed_decoder(string_table))
            object_arg_5: IRequiredGetter = get.Required
            return object_arg_5.Field("v", arg_11)

        return CompositeCell(1, _arrow2754()) if (match_value == "FreeText") else (CompositeCell(0, _arrow2755()) if (match_value == "Term") else (CompositeCell(2, _arrow2756(), _arrow2757()) if (match_value == "Unitized") else (CompositeCell(3, _arrow2758()) if (match_value == "Data") else to_fail(printf("Error reading CompositeCell from json string: %A"))(match_value))))

    return object(_arrow2759)


__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.arc_types import ArcWorkflow
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_iter)
from ...arctrl_core.Helper.identifier import (create_missing_identifier, Workflow_tryIdentifierFromFileName, Workflow_fileNameFromIdentifier)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.Process.component import Component
from ...fable_library.list import (of_array, FSharpList, map as map_1, empty, of_seq, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import default_arg
from ...fable_library.seq import map
from ...fable_library.seq2 import List_distinct
from ...fable_library.string_ import (starts_with_exact, join)
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .conversions import (OntologyAnnotation_fromAggregatedStrings, Component_fromAggregatedStrings, OntologyAnnotation_toAggregatedStrings, Component_toAggregatedStrings)
from .sparse_table import (SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

identifier_label: str = "Identifier"

title_label: str = "Title"

description_label: str = "Description"

workflow_type_label: str = "Type"

type_term_accession_number_label: str = "Type Term Accession Number"

type_term_source_reflabel: str = "Type Term Source REF"

sub_workflow_identifiers_label: str = "Sub Workflow Identifiers"

uri_label: str = "URI"

version_label: str = "Version"

parameters_name_label: str = "Parameters Name"

parameters_term_accession_number_label: str = "Parameters Term Accession Number"

parameters_term_source_reflabel: str = "Parameters Term Source REF"

components_name_label: str = "Components Name"

components_type_label: str = "Components Type"

components_type_term_accession_number_label: str = "Components Type Term Accession Number"

components_type_term_source_reflabel: str = "Components Type Term Source REF"

file_name_label: str = "File Name"

labels: FSharpList[str] = of_array([identifier_label, title_label, description_label, workflow_type_label, type_term_accession_number_label, type_term_source_reflabel, sub_workflow_identifiers_label, uri_label, version_label, parameters_name_label, parameters_term_accession_number_label, parameters_term_source_reflabel, components_name_label, components_type_label, components_type_term_accession_number_label, components_type_term_source_reflabel, file_name_label])

def from_string(identifier: str | None, title: str | None, description: str | None, workflow_type: str | None, workflow_type_term_accession_number: str | None, workflow_type_term_source_ref: str | None, subworkflow_identifiers: str | None, uri: str | None, version: str | None, parameters_name: str, parameters_term_accession_number: str, parameters_term_source_ref: str, components_name: str, components_type: str, components_type_term_accession_number: str, components_type_term_source_ref: str, file_name: str | None, comments: Array[Comment]) -> ArcWorkflow:
    subworkflow_identifiers_2: Array[str]
    if subworkflow_identifiers is None:
        subworkflow_identifiers_2 = []

    else: 
        subworkflow_identifiers_1: str = subworkflow_identifiers
        def mapping(s: str, identifier: Any=identifier, title: Any=title, description: Any=description, workflow_type: Any=workflow_type, workflow_type_term_accession_number: Any=workflow_type_term_accession_number, workflow_type_term_source_ref: Any=workflow_type_term_source_ref, subworkflow_identifiers: Any=subworkflow_identifiers, uri: Any=uri, version: Any=version, parameters_name: Any=parameters_name, parameters_term_accession_number: Any=parameters_term_accession_number, parameters_term_source_ref: Any=parameters_term_source_ref, components_name: Any=components_name, components_type: Any=components_type, components_type_term_accession_number: Any=components_type_term_accession_number, components_type_term_source_ref: Any=components_type_term_source_ref, file_name: Any=file_name, comments: Any=comments) -> str:
            return s.strip()

        subworkflow_identifiers_2 = list(map(mapping, subworkflow_identifiers_1.split(";")))

    workflow_type_1: OntologyAnnotation | None
    v: OntologyAnnotation = OntologyAnnotation.create(workflow_type, workflow_type_term_source_ref, workflow_type_term_accession_number)
    workflow_type_1 = Option_fromValueWithDefault(OntologyAnnotation(), v)
    parameters: Array[OntologyAnnotation] = list(OntologyAnnotation_fromAggregatedStrings(";", parameters_name, parameters_term_source_ref, parameters_term_accession_number))
    components: Array[Component] = list(Component_fromAggregatedStrings(";", components_name, components_type, components_type_term_source_ref, components_type_term_accession_number))
    identifier_4: str
    if identifier is None:
        if file_name is None:
            identifier_4 = create_missing_identifier()

        else: 
            match_value: str | None = Workflow_tryIdentifierFromFileName(file_name)
            identifier_4 = match_value if (match_value is not None) else create_missing_identifier()


    else: 
        identifier_4 = identifier

    return ArcWorkflow.make(identifier_4, title, description, workflow_type_1, uri, version, subworkflow_identifiers_2, parameters, components, None, [], None, comments)


def from_sparse_table(matrix: SparseTable) -> ArcWorkflow:
    def mapping(k: str, matrix: Any=matrix) -> Comment:
        return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, 0)))

    comments: FSharpList[Comment] = map_1(mapping, matrix.CommentKeys)
    return from_string(SparseTable__TryGetValue_11FD62A8(matrix, (identifier_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (title_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (description_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (workflow_type_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_accession_number_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (type_term_source_reflabel, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (sub_workflow_identifiers_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (uri_label, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (version_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_name_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_term_accession_number_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (parameters_term_source_reflabel, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_name_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_term_accession_number_label, 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (components_type_term_source_reflabel, 0)), SparseTable__TryGetValue_11FD62A8(matrix, (file_name_label, 0)), list(comments))


def to_sparse_table(workflow: ArcWorkflow) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, 2)
    comment_keys: FSharpList[str] = empty()
    pattern_input: tuple[str, str] = (("", "")) if starts_with_exact(workflow.Identifier, "MISSING_IDENTIFIER_") else ((workflow.Identifier, Workflow_fileNameFromIdentifier(workflow.Identifier)))
    wt: dict[str, Any]
    tt: OntologyAnnotation = default_arg(workflow.WorkflowType, OntologyAnnotation())
    wt = OntologyAnnotation.to_string_object(tt, True)
    p_agg: dict[str, Any] = OntologyAnnotation_toAggregatedStrings(";", list(workflow.Parameters))
    c_agg: dict[str, Any] = Component_toAggregatedStrings(";", of_seq(workflow.Components))
    sub_workflows_agg: str = join(";", workflow.SubWorkflowIdentifiers)
    add_to_dict(matrix.Matrix, (identifier_label, 1), pattern_input[0])
    add_to_dict(matrix.Matrix, (title_label, 1), default_arg(workflow.Title, ""))
    add_to_dict(matrix.Matrix, (description_label, 1), default_arg(workflow.Description, ""))
    add_to_dict(matrix.Matrix, (workflow_type_label, 1), wt["TermName"])
    add_to_dict(matrix.Matrix, (type_term_accession_number_label, 1), wt["TermAccessionNumber"])
    add_to_dict(matrix.Matrix, (type_term_source_reflabel, 1), wt["TermSourceREF"])
    add_to_dict(matrix.Matrix, (sub_workflow_identifiers_label, 1), sub_workflows_agg)
    add_to_dict(matrix.Matrix, (uri_label, 1), default_arg(workflow.URI, ""))
    add_to_dict(matrix.Matrix, (version_label, 1), default_arg(workflow.Version, ""))
    add_to_dict(matrix.Matrix, (parameters_name_label, 1), p_agg["TermNameAgg"])
    add_to_dict(matrix.Matrix, (parameters_term_accession_number_label, 1), p_agg["TermAccessionNumberAgg"])
    add_to_dict(matrix.Matrix, (parameters_term_source_reflabel, 1), p_agg["TermSourceREFAgg"])
    add_to_dict(matrix.Matrix, (components_name_label, 1), c_agg["NameAgg"])
    add_to_dict(matrix.Matrix, (components_type_label, 1), c_agg["TermNameAgg"])
    add_to_dict(matrix.Matrix, (components_type_term_accession_number_label, 1), c_agg["TermAccessionNumberAgg"])
    add_to_dict(matrix.Matrix, (components_type_term_source_reflabel, 1), c_agg["TermSourceREFAgg"])
    add_to_dict(matrix.Matrix, (file_name_label, 1), pattern_input[1])
    def f(comment: Comment, workflow: Any=workflow) -> None:
        nonlocal comment_keys
        pattern_input_1: tuple[str, str] = Comment_toString(comment)
        n: str = pattern_input_1[0]
        comment_keys = cons(n, comment_keys)
        add_to_dict(matrix.Matrix, (n, 1), pattern_input_1[1])

    ResizeArray_iter(f, workflow.Comments)
    class ObjectExpr3485:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3484(x: str, y: str) -> bool:
                return x == y

            return _arrow3484

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3485())), matrix.ColumnCount)


def from_rows(line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], ArcWorkflow]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number, "Workflow")
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(workflow: ArcWorkflow) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(to_sparse_table(workflow), "Workflow")


__all__ = ["identifier_label", "title_label", "description_label", "workflow_type_label", "type_term_accession_number_label", "type_term_source_reflabel", "sub_workflow_identifiers_label", "uri_label", "version_label", "parameters_name_label", "parameters_term_accession_number_label", "parameters_term_source_reflabel", "components_name_label", "components_type_label", "components_type_term_accession_number_label", "components_type_term_source_reflabel", "file_name_label", "labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from .arc_types import (ArcAssay, ArcStudy, ArcWorkflow, ArcRun, ArcInvestigation)
from .Helper.identifier import check_valid_characters
from .Table.arc_table import ArcTable

def set_arc_table_name(new_name: str, table: ArcTable) -> ArcTable:
    check_valid_characters(new_name)
    table.Name = new_name
    return table


def set_assay_identifier(new_identifier: str, assay: ArcAssay) -> ArcAssay:
    check_valid_characters(new_identifier)
    assay.Identifier = new_identifier
    return assay


def set_study_identifier(new_identifier: str, study: ArcStudy) -> ArcStudy:
    check_valid_characters(new_identifier)
    study.Identifier = new_identifier
    return study


def set_workflow_identifier(new_identifier: str, workflow: ArcWorkflow) -> ArcWorkflow:
    check_valid_characters(new_identifier)
    workflow.Identifier = new_identifier
    return workflow


def set_run_identifier(new_identifier: str, run: ArcRun) -> ArcRun:
    check_valid_characters(new_identifier)
    run.Identifier = new_identifier
    return run


def set_investigation_identifier(new_identifier: str, investigation: ArcInvestigation) -> ArcInvestigation:
    check_valid_characters(new_identifier)
    investigation.Identifier = new_identifier
    return investigation


__all__ = ["set_arc_table_name", "set_assay_identifier", "set_study_identifier", "set_workflow_identifier", "set_run_identifier", "set_investigation_identifier"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from .map_util import get_item_from_dict, try_get_value
from .reflection import TypeInfo, class_type
from .resize_array import find_index
from .seq import concat, delay, iterate, iterate_indexed, map, to_array
from .string_ import format
from .types import Array, FSharpRef
from .util import (
    ICollection,
    IEnumerable_1,
    IEnumerator,
    IEqualityComparer_1,
    dispose,
    equals,
    get_enumerator,
    ignore,
    to_iterator,
)


_KEY = TypeVar("_KEY")

_VALUE = TypeVar("_VALUE")


def _expr30(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return class_type("Fable.Collections.Dictionary", [gen0, gen1], Dictionary)


class Dictionary(Generic[_KEY, _VALUE]):
    def __init__(self, pairs: IEnumerable_1[Any], comparer: IEqualityComparer_1[Any]) -> None:
        this: FSharpRef[Dictionary[_KEY, _VALUE]] = FSharpRef(None)
        self.comparer: IEqualityComparer_1[Any] = comparer
        this.contents = self
        self.hash_map: Any = dict([])
        self.init_00409: int = 1
        with get_enumerator(pairs) as enumerator:
            while enumerator.System_Collections_IEnumerator_MoveNext():
                pair: Any = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                Dictionary__Add_5BDDA1(this.contents, pair[0], pair[1])

    @property
    def Symbol_toStringTag(self, __unit: None = None) -> str:
        return "Dictionary"

    def to_json(self, __unit: None = None) -> Any:
        this: Dictionary[_KEY, _VALUE] = self
        return list(this)

    def System_Collections_IEnumerable_GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        this: Dictionary[_KEY, _VALUE] = self
        return get_enumerator(this)

    def GetEnumerator(self, __unit: None = None) -> IEnumerator[Any]:
        this: Dictionary[_KEY, _VALUE] = self
        return get_enumerator(concat(this.hash_map.values()))

    def __iter__(self) -> IEnumerator[Any]:
        return to_iterator(self.GetEnumerator())

    def System_Collections_Generic_ICollection_1_Add2B595(self, item: Any) -> None:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__Add_5BDDA1(this, item[0], item[1])

    def System_Collections_Generic_ICollection_1_Clear(self, __unit: None = None) -> None:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__Clear(this)

    def System_Collections_Generic_ICollection_1_Contains2B595(self, item: Any) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        match_value: Any | None = Dictionary__TryFind_2B595(this, item[0])
        (pattern_matching_result,) = (None,)
        if match_value is not None:
            if equals(match_value[1], item[1]):
                pattern_matching_result = 0

            else:
                pattern_matching_result = 1

        else:
            pattern_matching_result = 1

        if pattern_matching_result == 0:
            return True

        elif pattern_matching_result == 1:
            return False

    def System_Collections_Generic_ICollection_1_CopyToZ3B4C077E(self, array: Array[Any], array_index: int) -> None:
        this: Dictionary[_KEY, _VALUE] = self

        def action(i: int, e: Any) -> None:
            array[array_index + i] = e

        iterate_indexed(action, this)

    def __len__(self, __unit: None = None) -> int:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__get_Count(this)

    def System_Collections_Generic_ICollection_1_get_IsReadOnly(self, __unit: None = None) -> bool:
        return False

    def System_Collections_Generic_ICollection_1_Remove2B595(self, item: Any) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        match_value: Any | None = Dictionary__TryFind_2B595(this, item[0])
        if match_value is not None:
            if equals(match_value[1], item[1]):
                ignore(Dictionary__Remove_2B595(this, item[0]))

            return True

        else:
            return False

    def System_Collections_Generic_IDictionary_2_Add5BDDA1(self, key: _KEY, value: _VALUE) -> None:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__Add_5BDDA1(this, key, value)

    def System_Collections_Generic_IDictionary_2_ContainsKey2B595(self, key: _KEY | None = None) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__ContainsKey_2B595(this, key)

    def System_Collections_Generic_IDictionary_2_get_Item2B595(self, key: _KEY | None = None) -> _VALUE:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__get_Item_2B595(this, key)

    def System_Collections_Generic_IDictionary_2_set_Item5BDDA1(self, key: _KEY, v: _VALUE) -> None:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__set_Item_5BDDA1(this, key, v)

    def System_Collections_Generic_IDictionary_2_get_Keys(self, __unit: None = None) -> ICollection[_KEY]:
        this: Dictionary[_KEY, _VALUE] = self

        def _arrow27(__unit: None = None) -> IEnumerable_1[_KEY]:
            def _arrow26(pair: Any) -> _KEY:
                return pair[0]

            return map(_arrow26, this)

        return to_array(delay(_arrow27))

    def System_Collections_Generic_IDictionary_2_Remove2B595(self, key: _KEY | None = None) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__Remove_2B595(this, key)

    def System_Collections_Generic_IDictionary_2_TryGetValue6DC89625(self, key: _KEY, value: FSharpRef[_VALUE]) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        match_value: Any | None = Dictionary__TryFind_2B595(this, key)
        if match_value is not None:
            pair: Any = match_value
            value.contents = pair[1]
            return True

        else:
            return False

    def System_Collections_Generic_IDictionary_2_get_Values(self, __unit: None = None) -> ICollection[_VALUE]:
        this: Dictionary[_KEY, _VALUE] = self

        def _arrow29(__unit: None = None) -> IEnumerable_1[_VALUE]:
            def _arrow28(pair: Any) -> _VALUE:
                return pair[1]

            return map(_arrow28, this)

        return to_array(delay(_arrow29))

    @property
    def size(self, __unit: None = None) -> int:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__get_Count(this)

    def clear(self, __unit: None = None) -> None:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__Clear(this)

    def __delitem__(self, k: _KEY | None = None) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__Remove_2B595(this, k)

    def entries(self, __unit: None = None) -> IEnumerable_1[tuple[_KEY, _VALUE]]:
        this: Dictionary[_KEY, _VALUE] = self

        def mapping(p: Any) -> tuple[_KEY, _VALUE]:
            return (p[0], p[1])

        return map(mapping, this)

    def __getitem__(self, k: _KEY | None = None) -> _VALUE:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__get_Item_2B595(this, k)

    def __contains__(self, k: _KEY | None = None) -> bool:
        this: Dictionary[_KEY, _VALUE] = self
        return Dictionary__ContainsKey_2B595(this, k)

    def keys(self, __unit: None = None) -> IEnumerable_1[_KEY]:
        this: Dictionary[_KEY, _VALUE] = self

        def mapping(p: Any) -> _KEY:
            return p[0]

        return map(mapping, this)

    def __setitem__(self, k: _KEY, v: _VALUE) -> Map_2[_KEY, _VALUE]:
        this: Dictionary[_KEY, _VALUE] = self
        Dictionary__set_Item_5BDDA1(this, k, v)
        return this

    def values(self, __unit: None = None) -> IEnumerable_1[_VALUE]:
        this: Dictionary[_KEY, _VALUE] = self

        def mapping(p: Any) -> _VALUE:
            return p[1]

        return map(mapping, this)

    def for_each(self, f: Callable[[_VALUE, _KEY, Map_2[_KEY, _VALUE]], None], this_arg: Any | None = None) -> None:
        this: Dictionary[_KEY, _VALUE] = self

        def action(p: Any) -> None:
            f(p[1], p[0], this)

        iterate(action, this)


Dictionary_reflection = _expr30


def Dictionary__ctor_6623D9B3(
    pairs: IEnumerable_1[Any], comparer: IEqualityComparer_1[Any]
) -> Dictionary[_KEY, _VALUE]:
    return Dictionary(pairs, comparer)


def Dictionary__TryFindIndex_2B595(this: Dictionary[Any, Any], k: Any) -> tuple[bool, int, int]:
    h: int = this.comparer.GetHashCode(k) or 0
    match_value: tuple[bool, Array[Any]]
    out_arg: Array[Any] = None

    def _arrow31(__unit: None = None, this: Any = this, k: Any = k) -> Array[Any]:
        return out_arg

    def _arrow32(v: Array[Any], this: Any = this, k: Any = k) -> None:
        nonlocal out_arg
        out_arg = v

    match_value = (try_get_value(this.hash_map, h, FSharpRef(_arrow31, _arrow32)), out_arg)
    if match_value[0]:

        def _arrow33(pair: Any, this: Any = this, k: Any = k) -> bool:
            return this.comparer.Equals(k, pair[0])

        return (True, h, find_index(_arrow33, match_value[1]))

    else:
        return (False, h, -1)


def Dictionary__TryFind_2B595(this: Dictionary[Any, Any], k: Any) -> Any | None:
    match_value: tuple[bool, int, int] = Dictionary__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return get_item_from_dict(this.hash_map, match_value[1])[match_value[2]]

    elif pattern_matching_result == 1:
        return None


def Dictionary__get_Comparer(this: Dictionary[Any, Any]) -> IEqualityComparer_1[Any]:
    return this.comparer


def Dictionary__Clear(this: Dictionary[Any, Any]) -> None:
    this.hash_map.clear()


def Dictionary__get_Count(this: Dictionary[Any, Any]) -> int:
    count: int = 0
    enumerator: Any = get_enumerator(this.hash_map.values())
    try:
        while enumerator.System_Collections_IEnumerator_MoveNext():
            pairs: Array[Any] = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
            count = (count + len(pairs)) or 0

    finally:
        dispose(enumerator)

    return count


def Dictionary__get_Item_2B595(this: Dictionary[Any, Any], k: Any) -> Any:
    match_value: Any | None = Dictionary__TryFind_2B595(this, k)
    if match_value is not None:
        return match_value[1]

    else:
        raise Exception("The item was not found in collection")


def Dictionary__set_Item_5BDDA1(this: Dictionary[Any, Any], k: Any, v: Any) -> None:
    match_value: tuple[bool, int, int] = Dictionary__TryFindIndex_2B595(this, k)
    if match_value[0]:
        if match_value[2] > -1:
            get_item_from_dict(this.hash_map, match_value[1])[match_value[2]] = (k, v)

        else:
            value: None = get_item_from_dict(this.hash_map, match_value[1]).append((k, v))
            ignore(None)

    else:
        this.hash_map[match_value[1]] = [(k, v)]


def Dictionary__Add_5BDDA1(this: Dictionary[Any, Any], k: Any, v: Any) -> None:
    match_value: tuple[bool, int, int] = Dictionary__TryFindIndex_2B595(this, k)
    if match_value[0]:
        if match_value[2] > -1:
            raise Exception(format("An item with the same key has already been added. Key: {0}", k))

        else:
            value: None = get_item_from_dict(this.hash_map, match_value[1]).append((k, v))
            ignore(None)

    else:
        this.hash_map[match_value[1]] = [(k, v)]


def Dictionary__ContainsKey_2B595(this: Dictionary[Any, Any], k: Any) -> bool:
    match_value: tuple[bool, int, int] = Dictionary__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return True

    elif pattern_matching_result == 1:
        return False


def Dictionary__Remove_2B595(this: Dictionary[Any, Any], k: Any) -> bool:
    match_value: tuple[bool, int, int] = Dictionary__TryFindIndex_2B595(this, k)
    (pattern_matching_result,) = (None,)
    if match_value[0]:
        if match_value[2] > -1:
            pattern_matching_result = 0

        else:
            pattern_matching_result = 1

    else:
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        get_item_from_dict(this.hash_map, match_value[1]).pop(match_value[2])
        return True

    elif pattern_matching_result == 1:
        return False


__all__ = [
    "Dictionary_reflection",
    "Dictionary__TryFindIndex_2B595",
    "Dictionary__TryFind_2B595",
    "Dictionary__get_Comparer",
    "Dictionary__Clear",
    "Dictionary__get_Count",
    "Dictionary__get_Item_2B595",
    "Dictionary__set_Item_5BDDA1",
    "Dictionary__Add_5BDDA1",
    "Dictionary__ContainsKey_2B595",
    "Dictionary__Remove_2B595",
]



from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...arctrl_core.comment import (Comment, Remark)
from ...arctrl_core.Helper.collections_ import ResizeArray_iter
from ...arctrl_core.ontology_source_reference import OntologySourceReference
from ...fable_library.list import (of_array, FSharpList, length, singleton, initialize, map, empty, iterate_indexed, cons, reverse)
from ...fable_library.map_util import add_to_dict
from ...fable_library.option import default_arg
from ...fable_library.seq2 import List_distinct
from ...fable_library.types import Array
from ...fable_library.util import (string_hash, IEnumerable_1, IEnumerator)
from .comment import (Comment_fromString, Comment_toString)
from .sparse_table import (SparseTable_GetEmptyComments_3ECCA699, SparseTable__TryGetValueDefault_5BAE6133, SparseTable__TryGetValue_11FD62A8, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1)

name_label: str = "Term Source Name"

file_label: str = "Term Source File"

version_label: str = "Term Source Version"

description_label: str = "Term Source Description"

labels: FSharpList[str] = of_array([name_label, file_label, version_label, description_label])

def from_string(description: str | None, file: str | None, name: str | None, version: str | None, comments: Array[Comment]) -> OntologySourceReference:
    return OntologySourceReference.make(description, file, name, version, comments)


def from_sparse_table(matrix: SparseTable) -> FSharpList[OntologySourceReference]:
    if (length(matrix.CommentKeys) != 0) if (matrix.ColumnCount == 0) else False:
        comments: Array[Comment] = SparseTable_GetEmptyComments_3ECCA699(matrix)
        def _arrow3489(__unit: None=None, matrix: Any=matrix) -> OntologySourceReference:
            return_val: OntologySourceReference = OntologySourceReference.create()
            return_val.Comments = comments
            return return_val

        return singleton(_arrow3489())

    else: 
        def _arrow3490(i: int, matrix: Any=matrix) -> OntologySourceReference:
            def mapping(k: str) -> Comment:
                return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, i)))

            comments_1: Array[Comment] = list(map(mapping, matrix.CommentKeys))
            return from_string(SparseTable__TryGetValue_11FD62A8(matrix, (description_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (file_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (name_label, i)), SparseTable__TryGetValue_11FD62A8(matrix, (version_label, i)), comments_1)

        return initialize(matrix.ColumnCount, _arrow3490)



def to_sparse_table(ontology_sources: FSharpList[OntologySourceReference]) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, labels, None, length(ontology_sources) + 1)
    comment_keys: FSharpList[str] = empty()
    def action(i: int, o: OntologySourceReference, ontology_sources: Any=ontology_sources) -> None:
        i_1: int = (i + 1) or 0
        add_to_dict(matrix.Matrix, (name_label, i_1), default_arg(o.Name, ""))
        add_to_dict(matrix.Matrix, (file_label, i_1), default_arg(o.File, ""))
        add_to_dict(matrix.Matrix, (version_label, i_1), default_arg(o.Version, ""))
        add_to_dict(matrix.Matrix, (description_label, i_1), default_arg(o.Description, ""))
        def f(comment: Comment, i: Any=i, o: Any=o) -> None:
            nonlocal comment_keys
            pattern_input: tuple[str, str] = Comment_toString(comment)
            n: str = pattern_input[0]
            comment_keys = cons(n, comment_keys)
            add_to_dict(matrix.Matrix, (n, i_1), pattern_input[1])

        ResizeArray_iter(f, o.Comments)

    iterate_indexed(action, ontology_sources)
    class ObjectExpr3492:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3491(x: str, y: str) -> bool:
                return x == y

            return _arrow3491

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3492())), matrix.ColumnCount)


def from_rows(line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[OntologySourceReference]]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, labels, line_number)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], from_sparse_table(tupled_arg[3]))


def to_rows(term_sources: FSharpList[OntologySourceReference]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(to_sparse_table(term_sources))


__all__ = ["name_label", "file_label", "version_label", "description_label", "labels", "from_string", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, Protocol, TypeVar

from .types import Array


_T = TypeVar("_T")


class Cons_1(Protocol, Generic[_T]):
    @abstractmethod
    def Allocate(self, len: int) -> Array[_T]: ...


def Helpers_allocateArrayFromCons(cons: Cons_1[Any], len_1: int) -> Array[Any]:
    if cons is None:
        return (list)([None] * len_1)

    else:
        return cons([0] * len_1)


def Helpers_fillImpl(array: Array[Any], value: Any, start: int, count: int) -> Array[Any]:
    for i in range(0, (count - 1) + 1, 1):
        array[i + start] = value
    return array


def Helpers_spliceImpl(array: Array[Any], start: int, delete_count: int) -> Array[Any]:
    for _ in range(1, delete_count + 1, 1):
        array.pop(start)
    return array


def Helpers_indexOfImpl(array: Array[Any], item: Any, start: int) -> Any:
    try:
        return array.index(item, start)

    except Exception as ex:
        return -1


__all__ = ["Helpers_allocateArrayFromCons", "Helpers_fillImpl", "Helpers_spliceImpl", "Helpers_indexOfImpl"]



from __future__ import annotations
from collections.abc import Callable
from datetime import datetime
from decimal import Decimal
from typing import (Any, TypeVar)
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDValue, LDRef, LDNode)
from ...fable_library.array_ import fold
from ...fable_library.list import (FSharpList, is_empty, length, head, of_array)
from ...fable_library.option import value as value_12
from ...fable_library.result import FSharpResult_2
from ...fable_library.seq import (to_list, delay, map, enumerate_from_functions, append, singleton, empty, collect)
from ...fable_library.string_ import starts_with_exact
from ...fable_library.types import Array
from ...fable_library.util import (equals, is_iterable, get_enumerator, IEnumerator, IEnumerable_1, int32_to_string)
from ...thoth_json_core.decode import (Getters_2__ctor_Z4BE6C149, Getters_2, IGetters, string, IRequiredGetter, IOptionalGetter, Getters_2__get_Errors, map as map_1, one_of, int_1, decimal)
from ...thoth_json_core.encode import list_1
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1, ErrorReason_1, IDecoderHelpers_1)
from ..decode import (Decode_resizeArrayOrSingleton, Helpers_prependPath)
from ..encode import (date_time, resize_array_or_singleton)
from .ldcontext import (encoder as encoder_3, decoder as decoder_1)
from .ldref import (encoder as encoder_2, decoder as decoder_3)
from .ldvalue import (encoder as encoder_1, decoder as decoder_2)

__A_ = TypeVar("__A_")

def generic_encoder(obj: Any=None) -> IEncodable:
    if str(type(obj)) == "<class \'str\'>":
        class ObjectExpr1909(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
                return helpers.encode_string(obj)

        return ObjectExpr1909()

    elif str(type(obj)) == "<class \'int\'>":
        class ObjectExpr1910(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
                return helpers_1.encode_signed_integral_number(obj)

        return ObjectExpr1910()

    elif str(type(obj)) == "<class \'bool\'>":
        class ObjectExpr1911(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
                return helpers_2.encode_bool(obj)

        return ObjectExpr1911()

    elif str(type(obj)) == "<class \'float\'>":
        class ObjectExpr1912(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
                return helpers_3.encode_decimal_number(obj)

        return ObjectExpr1912()

    elif isinstance(obj, datetime):
        return date_time(obj)

    elif isinstance(obj, LDValue):
        return encoder_1(obj)

    elif isinstance(obj, LDRef):
        return encoder_2(obj)

    elif isinstance(obj, LDNode):
        return encoder(obj)

    elif equals(obj, None):
        class ObjectExpr1913(IEncodable):
            def Encode(self, helpers_4: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
                return helpers_4.encode_null()

        return ObjectExpr1913()

    elif is_iterable(obj):
        def _arrow1917(__unit: None=None, obj: Any=obj) -> IEnumerable_1[IEncodable]:
            def _arrow1914(__unit: None=None) -> IEnumerator[Any]:
                return get_enumerator(obj)

            def _arrow1915(enumerator: IEnumerator[Any]) -> bool:
                return enumerator.System_Collections_IEnumerator_MoveNext()

            def _arrow1916(enumerator_1: IEnumerator[Any]) -> Any:
                return enumerator_1.System_Collections_IEnumerator_get_Current()

            return map(generic_encoder, enumerate_from_functions(_arrow1914, _arrow1915, _arrow1916))

        return list_1(to_list(delay(_arrow1917)))

    else: 
        raise Exception("Unknown type")



def encoder(obj: LDNode) -> IEncodable:
    def _arrow1930(__unit: None=None, obj: Any=obj) -> IEnumerable_1[tuple[str, IEncodable]]:
        def _arrow1919(__unit: None=None) -> IEncodable:
            value: str = obj.Id
            class ObjectExpr1918(IEncodable):
                def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                    return helpers.encode_string(value)

            return ObjectExpr1918()

        def _arrow1929(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
            def _arrow1921(value_1: str) -> IEncodable:
                class ObjectExpr1920(IEncodable):
                    def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                        return helpers_1.encode_string(value_1)

                return ObjectExpr1920()

            def _arrow1928(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                def _arrow1923(value_3: str) -> IEncodable:
                    class ObjectExpr1922(IEncodable):
                        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                            return helpers_2.encode_string(value_3)

                    return ObjectExpr1922()

                def _arrow1927(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                    def _arrow1924(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                        match_value: LDContext | None = obj.TryGetContext()
                        if match_value is not None:
                            return singleton(("@context", encoder_3(match_value)))

                        else: 
                            return empty()


                    def _arrow1926(__unit: None=None) -> IEnumerable_1[tuple[str, IEncodable]]:
                        def _arrow1925(kv: Any) -> IEnumerable_1[tuple[str, IEncodable]]:
                            l: str = kv[0].lower()
                            return singleton((kv[0], generic_encoder(kv[1]))) if ((not starts_with_exact(l, "init_")) if ((not starts_with_exact(l, "init@")) if ((l != "@context") if ((l != "additionaltype") if ((l != "schematype") if (l != "id") else False) else False) else False) else False) else False) else empty()

                        return collect(_arrow1925, obj.GetProperties(True))

                    return append(_arrow1924(), delay(_arrow1926))

                return append(singleton(("additionalType", resize_array_or_singleton(_arrow1923, obj.AdditionalType))) if (len(obj.AdditionalType) != 0) else empty(), delay(_arrow1927))

            return append(singleton(("@type", resize_array_or_singleton(_arrow1921, obj.SchemaType))), delay(_arrow1928))

        return append(singleton(("@id", _arrow1919())), delay(_arrow1929))

    values: FSharpList[tuple[str, IEncodable]] = to_list(delay(_arrow1930))
    class ObjectExpr1931(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], obj: Any=obj) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr1931()


def get_decoder(expect_object: bool) -> Decoder_1[Any]:
    def decode(expect_object_1: bool, expect_object: Any=expect_object) -> Decoder_1[Any]:
        class ObjectExpr1934(Decoder_1[LDNode]):
            def Decode(self, helpers: IDecoderHelpers_1[Any], value: Any, expect_object_1: Any=expect_object_1) -> FSharpResult_2[LDNode, tuple[str, ErrorReason_1[__A_]]]:
                if helpers.is_object(value):
                    getters: Getters_2[__A_, Any] = Getters_2__ctor_Z4BE6C149(helpers, value)
                    properties: IEnumerable_1[str] = helpers.get_properties(value)
                    result: LDNode
                    get: IGetters = getters
                    t: Array[str]
                    arg_1: Decoder_1[Array[str]] = Decode_resizeArrayOrSingleton(string)
                    object_arg: IRequiredGetter = get.Required
                    t = object_arg.Field("@type", arg_1)
                    id: str
                    object_arg_1: IRequiredGetter = get.Required
                    id = object_arg_1.Field("@id", string)
                    context: LDContext | None
                    object_arg_2: IOptionalGetter = get.Optional
                    context = object_arg_2.Field("@context", decoder_1)
                    def _arrow1932(__unit: None=None) -> Array[str] | None:
                        arg_7: Decoder_1[Array[str]] = Decode_resizeArrayOrSingleton(string)
                        object_arg_3: IOptionalGetter = get.Optional
                        return object_arg_3.Field("additionalType", arg_7)

                    o: LDNode = LDNode(id, t, _arrow1932())
                    with get_enumerator(properties) as enumerator:
                        while enumerator.System_Collections_IEnumerator_MoveNext():
                            property: str = enumerator.System_Collections_Generic_IEnumerator_1_get_Current()
                            if (property != "@context") if ((property != "@type") if (property != "@id") else False) else False:
                                def _arrow1933(__unit: None=None) -> Any:
                                    arg_9: Decoder_1[Any] = decode(False)
                                    object_arg_4: IRequiredGetter = get.Required
                                    return object_arg_4.Field(property, arg_9)

                                o.SetProperty(property, _arrow1933())

                    if context is not None:
                        o.SetContext(value_12(context))

                    result = o
                    match_value: FSharpList[tuple[str, ErrorReason_1[__A_]]] = Getters_2__get_Errors(getters)
                    if not is_empty(match_value):
                        errors: FSharpList[tuple[str, ErrorReason_1[__A_]]] = match_value
                        return FSharpResult_2(1, ("", ErrorReason_1(7, errors))) if (length(errors) > 1) else FSharpResult_2(1, head(match_value))

                    else: 
                        return FSharpResult_2(0, result)


                else: 
                    return FSharpResult_2(1, ("", ErrorReason_1(0, "an object", value)))


        decode_object: Decoder_1[LDNode] = ObjectExpr1934()
        class ObjectExpr1936(Decoder_1[Array[Any]]):
            def Decode(self, helpers_1: IDecoderHelpers_1[Any], value_1: Any, expect_object_1: Any=expect_object_1) -> FSharpResult_2[Array[Any], tuple[str, ErrorReason_1[__A_]]]:
                if helpers_1.is_array(value_1):
                    i: int = -1
                    def folder(acc: FSharpResult_2[Array[Any], tuple[str, ErrorReason_1[__A_]]], value_2: __A_) -> FSharpResult_2[Array[Any], tuple[str, ErrorReason_1[__A_]]]:
                        nonlocal i
                        i = (i + 1) or 0
                        if acc.tag == 0:
                            acc_1: Array[Any] = acc.fields[0]
                            match_value_1: FSharpResult_2[Any, tuple[str, ErrorReason_1[__A_]]]
                            copy_of_struct: Decoder_1[Any] = decode(False)
                            match_value_1 = copy_of_struct.Decode(helpers_1, value_2)
                            if match_value_1.tag == 0:
                                (acc_1.append(match_value_1.fields[0]))
                                return FSharpResult_2(0, acc_1)

                            else: 
                                def _arrow1935(__unit: None=None, acc: Any=acc, value_2: Any=value_2) -> tuple[str, ErrorReason_1[__A_]]:
                                    tupled_arg: tuple[str, ErrorReason_1[__A_]] = match_value_1.fields[0]
                                    return Helpers_prependPath((".[" + int32_to_string(i)) + "]", tupled_arg[0], tupled_arg[1])

                                return FSharpResult_2(1, _arrow1935())


                        else: 
                            return acc


                    return fold(folder, FSharpResult_2(0, []), helpers_1.as_array(value_1))

                else: 
                    return FSharpResult_2(1, ("", ErrorReason_1(0, "an array", value_1)))


        resize_array: Decoder_1[Array[Any]] = ObjectExpr1936()
        if expect_object_1:
            def _arrow1937(value_4: LDNode, expect_object_1: Any=expect_object_1) -> Any:
                return value_4

            return map_1(_arrow1937, decode_object)

        else: 
            def _arrow1939(value_5: LDValue, expect_object_1: Any=expect_object_1) -> Any:
                return value_5

            def _arrow1940(value_6: LDNode, expect_object_1: Any=expect_object_1) -> Any:
                return value_6

            def _arrow1941(value_7: LDRef, expect_object_1: Any=expect_object_1) -> Any:
                return value_7

            def _arrow1942(value_8: Array[Any], expect_object_1: Any=expect_object_1) -> Any:
                return value_8

            def _arrow1943(value_9: str, expect_object_1: Any=expect_object_1) -> Any:
                return value_9

            def _arrow1944(value_10: int, expect_object_1: Any=expect_object_1) -> Any:
                return value_10

            def _arrow1945(value_11: Decimal, expect_object_1: Any=expect_object_1) -> Any:
                return value_11

            return one_of(of_array([map_1(_arrow1939, decoder_2), map_1(_arrow1940, decode_object), map_1(_arrow1941, decoder_3), map_1(_arrow1942, resize_array), map_1(_arrow1943, string), map_1(_arrow1944, int_1), map_1(_arrow1945, decimal)]))


    return decode(expect_object)


def _arrow1946(value: Any=None) -> LDNode:
    return value


decoder: Decoder_1[LDNode] = map_1(_arrow1946, get_decoder(True))

generic_decoder: Decoder_1[Any] = get_decoder(False)

__all__ = ["generic_encoder", "encoder", "get_decoder", "decoder", "generic_decoder"]




from __future__ import annotations
from collections.abc import Callable
from ..fable_library.array_ import (map, remove_in_place, try_find as try_find_1, collect)
from ..fable_library.option import (value as value_1, default_arg)
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.seq import (exists, iterate, try_item, try_find)
from ..fable_library.string_ import (to_fail, printf)
from ..fable_library.types import Array
from ..fable_library.util import (IEnumerable_1, ignore, equals, safe_hash, IDisposable)
from .fs_worksheet import FsWorksheet
from .Tables.fs_table import FsTable

def _expr254() -> TypeInfo:
    return class_type("FsSpreadsheet.FsWorkbook", None, FsWorkbook)


class FsWorkbook(IDisposable):
    def __init__(self, __unit: None=None) -> None:
        self._worksheets: Array[FsWorksheet] = []

    def Copy(self, __unit: None=None) -> FsWorkbook:
        self_1: FsWorkbook = self
        def mapping(s: FsWorksheet) -> FsWorksheet:
            return s.Copy()

        shts: Array[FsWorksheet] = map(mapping, self_1.GetWorksheets()[:], None)
        wb: FsWorkbook = FsWorkbook()
        wb.AddWorksheets(shts)
        return wb

    @staticmethod
    def copy(workbook: FsWorkbook) -> FsWorkbook:
        return workbook.Copy()

    def InitWorksheet(self, name: str) -> FsWorksheet:
        self_1: FsWorkbook = self
        sheet: FsWorksheet = FsWorksheet(name)
        (self_1._worksheets.append(sheet))
        return sheet

    @staticmethod
    def init_worksheet(name: str, workbook: FsWorkbook) -> FsWorksheet:
        return workbook.InitWorksheet(name)

    def AddWorksheet(self, sheet: FsWorksheet) -> None:
        self_1: FsWorkbook = self
        def predicate(ws: FsWorksheet) -> bool:
            return ws.Name == sheet.Name

        if exists(predicate, self_1._worksheets):
            arg: str = sheet.Name
            to_fail(printf("Could not add worksheet with name \"%s\" to workbook as it already contains a worksheet with the same name"))(arg)

        else: 
            (self_1._worksheets.append(sheet))


    @staticmethod
    def add_worksheet(sheet: FsWorksheet, workbook: FsWorkbook) -> FsWorkbook:
        workbook.AddWorksheet(sheet)
        return workbook

    def AddWorksheets(self, sheets: IEnumerable_1[FsWorksheet]) -> None:
        self_1: FsWorkbook = self
        def action(sheet: FsWorksheet) -> None:
            self_1.AddWorksheet(sheet)

        iterate(action, sheets)

    @staticmethod
    def add_worksheets(sheets: IEnumerable_1[FsWorksheet], workbook: FsWorkbook) -> FsWorkbook:
        workbook.AddWorksheets(sheets)
        return workbook

    def GetWorksheets(self, __unit: None=None) -> Array[FsWorksheet]:
        self_1: FsWorkbook = self
        return self_1._worksheets

    @staticmethod
    def get_worksheets(workbook: FsWorkbook) -> Array[FsWorksheet]:
        return workbook.GetWorksheets()

    def TryGetWorksheetAt(self, index: int) -> FsWorksheet | None:
        self_1: FsWorkbook = self
        return try_item(index - 1, self_1._worksheets)

    @staticmethod
    def try_get_worksheet_at(index: int, workbook: FsWorkbook) -> FsWorksheet | None:
        return workbook.TryGetWorksheetAt(index)

    def GetWorksheetAt(self, index: int) -> FsWorksheet:
        self_1: FsWorkbook = self
        match_value: FsWorksheet | None = self_1.TryGetWorksheetAt(index)
        if match_value is None:
            raise Exception(("FsWorksheet at position " + str(index)) + " is not present in the FsWorkbook.")

        else: 
            return match_value


    @staticmethod
    def get_worksheet_at(index: int, workbook: FsWorkbook) -> FsWorksheet:
        return workbook.GetWorksheetAt(index)

    def TryGetWorksheetByName(self, sheet_name: str) -> FsWorksheet | None:
        self_1: FsWorkbook = self
        def predicate(w: FsWorksheet) -> bool:
            return w.Name == sheet_name

        return try_find(predicate, self_1._worksheets)

    @staticmethod
    def try_get_worksheet_by_name(sheet_name: str, workbook: FsWorkbook) -> FsWorksheet | None:
        return workbook.TryGetWorksheetByName(sheet_name)

    def GetWorksheetByName(self, sheet_name: str) -> FsWorksheet:
        self_1: FsWorkbook = self
        try: 
            return value_1(self_1.TryGetWorksheetByName(sheet_name))

        except Exception as match_value:
            raise Exception(("FsWorksheet with name " + sheet_name) + " is not present in the FsWorkbook.")


    @staticmethod
    def get_worksheet_by_name(sheet_name: str, workbook: FsWorkbook) -> FsWorksheet:
        return workbook.GetWorksheetByName(sheet_name)

    def RemoveWorksheet(self, name: str) -> None:
        self_1: FsWorkbook = self
        def _arrow252(__unit: None=None) -> FsWorksheet:
            try: 
                def _arrow251(ws: FsWorksheet) -> bool:
                    return ws.Name == name

                return default_arg(try_find_1(_arrow251, self_1._worksheets), None)

            except Exception as match_value:
                raise Exception(("FsWorksheet with name " + name) + " was not found in FsWorkbook.")


        class ObjectExpr253:
            @property
            def Equals(self) -> Callable[[FsWorksheet, FsWorksheet], bool]:
                return equals

            @property
            def GetHashCode(self) -> Callable[[FsWorksheet], int]:
                return safe_hash

        ignore(remove_in_place(_arrow252(), self_1._worksheets, ObjectExpr253()))

    @staticmethod
    def remove_worksheet(name: str, workbook: FsWorkbook) -> FsWorkbook:
        workbook.RemoveWorksheet(name)
        return workbook

    def GetTables(self, __unit: None=None) -> Array[FsTable]:
        self_1: FsWorkbook = self
        def mapping(s: FsWorksheet) -> Array[FsTable]:
            return list(s.Tables)

        return collect(mapping, self_1.GetWorksheets()[:], None)

    @staticmethod
    def get_tables(workbook: FsWorkbook) -> Array[FsTable]:
        return workbook.GetTables()

    @staticmethod
    def validate_for_write(workbook: FsWorkbook) -> None:
        try: 
            def action(ws: FsWorksheet) -> None:
                FsWorksheet.validate_for_write(ws)

            iterate(action, workbook.GetWorksheets())

        except Exception as ex:
            raise Exception(("FsWorkbook could not be validated for write: " + str(ex)) + "")


    def Dispose(self, __unit: None=None) -> None:
        pass


FsWorkbook_reflection = _expr254

def FsWorkbook__ctor(__unit: None=None) -> FsWorkbook:
    return FsWorkbook(__unit)


__all__ = ["FsWorkbook_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import (value as value_1, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1578() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.Person", None, Person, LDNode_reflection())


class Person(LDNode):
    def __init__(self, id: str, given_name: Any=None, additional_type: Array[str] | None=None, family_name: Any | None=None, email: Any | None=None, identifier: Any | None=None, affiliation: Any | None=None, job_title: Any | None=None, additional_name: Any | None=None, address: Any | None=None, telephone: Any | None=None, fax_number: Any | None=None, disambiguating_description: Any | None=None) -> None:
        super().__init__(id, ["schema.org/Person"], default_arg(additional_type, []))
        this: FSharpRef[Person] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_property("givenName", given_name, this.contents)
        set_optional_property("familyName", family_name, this.contents)
        set_optional_property("email", email, this.contents)
        set_optional_property("identifier", identifier, this.contents)
        set_optional_property("affiliation", affiliation, this.contents)
        set_optional_property("jobTitle", job_title, this.contents)
        set_optional_property("additionalName", additional_name, this.contents)
        set_optional_property("address", address, this.contents)
        set_optional_property("telephone", telephone, this.contents)
        set_optional_property("faxNumber", fax_number, this.contents)
        set_optional_property("disambiguatingDescription", disambiguating_description, this.contents)

    def GetGivenName(self, __unit: None=None) -> str:
        this: Person = self
        obj: DynamicObj = this
        if try_get_property_value("givenName", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("givenName")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "givenName") + "\' is set on this \'") + "Person") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "givenName") + "\' set on this \'") + "Person") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_given_name() -> Callable[[Person], str]:
        def _arrow1577(p: Person) -> str:
            return p.GetGivenName()

        return _arrow1577


Person_reflection = _expr1578

def Person__ctor_16EB4AE1(id: str, given_name: Any=None, additional_type: Array[str] | None=None, family_name: Any | None=None, email: Any | None=None, identifier: Any | None=None, affiliation: Any | None=None, job_title: Any | None=None, additional_name: Any | None=None, address: Any | None=None, telephone: Any | None=None, fax_number: Any | None=None, disambiguating_description: Any | None=None) -> Person:
    return Person(id, given_name, additional_type, family_name, email, identifier, affiliation, job_title, additional_name, address, telephone, fax_number, disambiguating_description)


__all__ = ["Person_reflection"]




from __future__ import annotations
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
import requests
from typing import (Any, Protocol)
from ...fable_library.async_builder import (singleton as singleton_1, Async)
from ...fable_library.map import of_list
from ...fable_library.option import map
from ...fable_library.reflection import (TypeInfo, int32_type, string_type, class_type, record_type)
from ...fable_library.seq import (to_list, delay, map as map_1, collect, singleton)
from ...fable_library.types import Record
from ...fable_library.util import (create_obj, IEnumerable_1, compare_primitives)

class InteropResponseType(Protocol):
    @property
    @abstractmethod
    def encoding(self) -> str:
        ...

    @property
    @abstractmethod
    def headers(self) -> Any:
        ...

    @property
    @abstractmethod
    def status_code(self) -> int:
        ...

    @property
    @abstractmethod
    def text(self) -> str:
        ...


def _expr4025() -> TypeInfo:
    return record_type("ARCtrl.WebRequestHelpers.Py.Response", [], Response, lambda: [("status_code", int32_type), ("text", string_type), ("headers", class_type("Microsoft.FSharp.Collections.FSharpMap`2", [string_type, string_type])), ("encoding", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class Response(Record):
    status_code: int
    text: str
    headers: Any
    encoding: str

Response_reflection = _expr4025

def _expr4026() -> TypeInfo:
    return class_type("ARCtrl.WebRequestHelpers.Py.Requests", None, Requests)


class Requests:
    def __init__(self, __unit: None=None) -> None:
        pass


Requests_reflection = _expr4026

def Requests__ctor(__unit: None=None) -> Requests:
    return Requests(__unit)


def Requests_createHeadersDict_9F3777D(headers: Any | None=None) -> Any | None:
    def mapping(values: Any, headers: Any=headers) -> Any:
        def _arrow4028(__unit: None=None, values: Any=values) -> IEnumerable_1[tuple[str, Any]]:
            def _arrow4027(pair: Any) -> tuple[str, Any]:
                return (pair[0], pair[1])

            return map_1(_arrow4027, values)

        return create_obj(to_list(delay(_arrow4028)))

    return map(mapping, headers)


def Requests_mapResponseType_Z730CA7D4(response: InteropResponseType) -> Response:
    status_code: int = response.status_code or 0
    text: str = response.text
    encoding: str = response.text
    def _arrow4030(__unit: None=None, response: Any=response) -> IEnumerable_1[tuple[str, str]]:
        def _arrow4029(header_name: str) -> IEnumerable_1[tuple[str, str]]:
            return singleton((header_name, response.headers[header_name]))

        return collect(_arrow4029, list(response.headers.keys()))

    class ObjectExpr4031:
        @property
        def Compare(self) -> Callable[[str, str], int]:
            return compare_primitives

    return Response(status_code, text, of_list(to_list(delay(_arrow4030)), ObjectExpr4031()), encoding)


def Requests_get_43A074E6(url: str, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.get(url, headers=headers_dict))


def Requests_post_147AE53E(url: str, data: str | None=None, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.post(url, data=data, headers=headers_dict))


def Requests_put_147AE53E(url: str, data: str | None=None, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.put(url, data=data, headers=headers_dict))


def Requests_delete_147AE53E(url: str, data: str | None=None, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.delete(url, data=data, headers=headers_dict))


def Requests_head_147AE53E(url: str, data: str | None=None, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.head(url, data=data, headers=headers_dict))


def Requests_options_147AE53E(url: str, data: str | None=None, headers: Any | None=None) -> Response:
    headers_dict: Any | None = Requests_createHeadersDict_9F3777D(headers)
    return Requests_mapResponseType_Z730CA7D4(requests.head(url, data=data, headers=headers_dict))


def download_file(url: str) -> Async[str]:
    def _arrow4032(__unit: None=None, url: Any=url) -> Async[str]:
        response: Response = Requests_get_43A074E6(url)
        return singleton_1.Return(response.text)

    return singleton_1.Delay(_arrow4032)


__all__ = ["Response_reflection", "Requests_reflection", "Requests_createHeadersDict_9F3777D", "Requests_mapResponseType_Z730CA7D4", "Requests_get_43A074E6", "Requests_post_147AE53E", "Requests_put_147AE53E", "Requests_delete_147AE53E", "Requests_head_147AE53E", "Requests_options_147AE53E", "download_file"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ..arctrl_core.arc_types import (ArcInvestigation, ArcStudy, ArcAssay)
from ..arctrl_core.comment import (Comment, Comment_reflection, Remark)
from ..arctrl_core.Helper.collections_ import (ResizeArray_iter, Option_fromValueWithDefault)
from ..arctrl_core.ontology_source_reference import OntologySourceReference
from ..arctrl_core.person import Person
from ..arctrl_core.publication import Publication
from ..fable_library.list import (FSharpList, of_array, map, empty, cons, reverse, append, unzip, concat, of_seq, is_empty as is_empty_1, tail, head)
from ..fable_library.map import (of_list, try_find)
from ..fable_library.map_util import add_to_dict
from ..fable_library.option import (default_arg, value as value_1)
from ..fable_library.reflection import (TypeInfo, string_type, list_type, record_type)
from ..fable_library.seq import (is_empty, delay, append as append_1, singleton, collect, to_list, map as map_1, try_find as try_find_1, iterate_indexed)
from ..fable_library.seq2 import (List_distinct, List_distinctBy)
from ..fable_library.string_ import (to_fail, printf)
from ..fable_library.types import (Record, Array)
from ..fable_library.util import (string_hash, IEnumerable_1, IEnumerator, get_enumerator, ignore, to_enumerable, compare_primitives)
from ..fs_spreadsheet.fs_row import FsRow
from ..fs_spreadsheet.fs_workbook import FsWorkbook
from ..fs_spreadsheet.fs_worksheet import FsWorksheet
from .Metadata.comment import (Comment_fromString, Comment_toString, Remark_wrapRemark)
from .Metadata.contacts import (from_rows as from_rows_2, to_rows as to_rows_2)
from .Metadata.ontology_source_reference import (from_rows, to_rows)
from .Metadata.publication import (from_rows as from_rows_1, to_rows as to_rows_1)
from .Metadata.sparse_table import (SparseTable__TryGetValueDefault_5BAE6133, SparseTable, SparseTable_Create_Z2192E64B, SparseTable_FromRows_Z5579EC29, SparseTable_ToRows_759CAFC1, SparseRowModule_tryGetValueAt, SparseRowModule_fromValues, SparseRowModule_getAllValues, SparseRowModule_fromAllValues, SparseRowModule_fromFsRow, SparseRowModule_writeToSheet)
from .Metadata.study import (from_rows as from_rows_3, to_rows as to_rows_3)

def _expr3649() -> TypeInfo:
    return record_type("ARCtrl.Spreadsheet.ArcInvestigation.InvestigationInfo", [], ArcInvestigation_InvestigationInfo, lambda: [("Identifier", string_type), ("Title", string_type), ("Description", string_type), ("SubmissionDate", string_type), ("PublicReleaseDate", string_type), ("Comments", list_type(Comment_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class ArcInvestigation_InvestigationInfo(Record):
    Identifier: str
    Title: str
    Description: str
    SubmissionDate: str
    PublicReleaseDate: str
    Comments: FSharpList[Comment]

ArcInvestigation_InvestigationInfo_reflection = _expr3649

def ArcInvestigation_InvestigationInfo_create(identifier: str, title: str, description: str, submission_date: str, public_release_date: str, comments: FSharpList[Comment]) -> ArcInvestigation_InvestigationInfo:
    return ArcInvestigation_InvestigationInfo(identifier, title, description, submission_date, public_release_date, comments)


def ArcInvestigation_InvestigationInfo_get_Labels(__unit: None=None) -> FSharpList[str]:
    return of_array(["Investigation Identifier", "Investigation Title", "Investigation Description", "Investigation Submission Date", "Investigation Public Release Date"])


def ArcInvestigation_InvestigationInfo_FromSparseTable_3ECCA699(matrix: SparseTable) -> ArcInvestigation_InvestigationInfo:
    def mapping(k: str, matrix: Any=matrix) -> Comment:
        return Comment_fromString(k, SparseTable__TryGetValueDefault_5BAE6133(matrix, "", (k, 0)))

    comments: FSharpList[Comment] = map(mapping, matrix.CommentKeys)
    return ArcInvestigation_InvestigationInfo_create(SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Investigation Identifier", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Investigation Title", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Investigation Description", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Investigation Submission Date", 0)), SparseTable__TryGetValueDefault_5BAE6133(matrix, "", ("Investigation Public Release Date", 0)), comments)


def ArcInvestigation_InvestigationInfo_ToSparseTable_Z720BD3FF(investigation: ArcInvestigation) -> SparseTable:
    matrix: SparseTable = SparseTable_Create_Z2192E64B(None, ArcInvestigation_InvestigationInfo_get_Labels(), None, 2)
    comment_keys: FSharpList[str] = empty()
    add_to_dict(matrix.Matrix, ("Investigation Identifier", 1), investigation.Identifier)
    add_to_dict(matrix.Matrix, ("Investigation Title", 1), default_arg(investigation.Title, ""))
    add_to_dict(matrix.Matrix, ("Investigation Description", 1), default_arg(investigation.Description, ""))
    add_to_dict(matrix.Matrix, ("Investigation Submission Date", 1), default_arg(investigation.SubmissionDate, ""))
    add_to_dict(matrix.Matrix, ("Investigation Public Release Date", 1), default_arg(investigation.PublicReleaseDate, ""))
    def f(comment: Comment, investigation: Any=investigation) -> None:
        nonlocal comment_keys
        pattern_input: tuple[str, str] = Comment_toString(comment)
        n: str = pattern_input[0]
        comment_keys = cons(n, comment_keys)
        add_to_dict(matrix.Matrix, (n, 1), pattern_input[1])

    ResizeArray_iter(f, investigation.Comments)
    class ObjectExpr3651:
        @property
        def Equals(self) -> Callable[[str, str], bool]:
            def _arrow3650(x: str, y: str) -> bool:
                return x == y

            return _arrow3650

        @property
        def GetHashCode(self) -> Callable[[str], int]:
            return string_hash

    return SparseTable(matrix.Matrix, matrix.Keys, reverse(List_distinct(comment_keys, ObjectExpr3651())), matrix.ColumnCount)


def ArcInvestigation_InvestigationInfo_fromRows(line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], ArcInvestigation_InvestigationInfo]:
    tupled_arg: tuple[str | None, int, FSharpList[Remark], SparseTable] = SparseTable_FromRows_Z5579EC29(rows, ArcInvestigation_InvestigationInfo_get_Labels(), line_number)
    return (tupled_arg[0], tupled_arg[1], tupled_arg[2], ArcInvestigation_InvestigationInfo_FromSparseTable_3ECCA699(tupled_arg[3]))


def ArcInvestigation_InvestigationInfo_toRows_Z720BD3FF(investigation: ArcInvestigation) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return SparseTable_ToRows_759CAFC1(ArcInvestigation_InvestigationInfo_ToSparseTable_Z720BD3FF(investigation))


def ArcInvestigation_fromParts(investigation_info: ArcInvestigation_InvestigationInfo, ontology_source_reference: FSharpList[OntologySourceReference], publications: FSharpList[Publication], contacts: FSharpList[Person], studies: FSharpList[ArcStudy], assays: FSharpList[ArcAssay], remarks: FSharpList[Remark]) -> ArcInvestigation:
    def mapping(s: ArcStudy, investigation_info: Any=investigation_info, ontology_source_reference: Any=ontology_source_reference, publications: Any=publications, contacts: Any=contacts, studies: Any=studies, assays: Any=assays, remarks: Any=remarks) -> str:
        return s.Identifier

    study_identifiers: FSharpList[str] = map(mapping, studies)
    title: str | None = Option_fromValueWithDefault("", investigation_info.Title)
    description: str | None = Option_fromValueWithDefault("", investigation_info.Description)
    submission_date: str | None = Option_fromValueWithDefault("", investigation_info.SubmissionDate)
    public_release_date: str | None = Option_fromValueWithDefault("", investigation_info.PublicReleaseDate)
    ontology_source_references: Array[OntologySourceReference] = list(ontology_source_reference)
    publications_1: Array[Publication] = list(publications)
    contacts_1: Array[Person] = list(contacts)
    assays_1: Array[ArcAssay] = list(assays)
    studies_1: Array[ArcStudy] = list(studies)
    registered_study_identifiers: Array[str] = list(study_identifiers)
    comments: Array[Comment] = list(investigation_info.Comments)
    remarks_1: Array[Remark] = list(remarks)
    return ArcInvestigation.make(investigation_info.Identifier, title, description, submission_date, public_release_date, ontology_source_references, publications_1, contacts_1, assays_1, studies_1, [], [], registered_study_identifiers, comments, remarks_1)


def ArcInvestigation_fromRows(rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]]) -> ArcInvestigation:
    if is_empty(rows):
        raise Exception("isa_investigation sheet in Investigation file is empty")

    en: IEnumerator[IEnumerable_1[tuple[int, str]]] = get_enumerator(rows)
    empty_investigation_info: ArcInvestigation_InvestigationInfo = ArcInvestigation_InvestigationInfo_create("", "", "", "", "", empty())
    def loop(last_line_mut: str | None, ontology_source_references_mut: FSharpList[OntologySourceReference], investigation_info_mut: ArcInvestigation_InvestigationInfo, publications_mut: FSharpList[Publication], contacts_mut: FSharpList[Person], studies_mut: FSharpList[tuple[ArcStudy, FSharpList[ArcAssay]]], remarks_mut: FSharpList[Remark], line_number_mut: int, rows: Any=rows) -> ArcInvestigation:
        while True:
            (last_line, ontology_source_references, investigation_info, publications, contacts, studies, remarks, line_number) = (last_line_mut, ontology_source_references_mut, investigation_info_mut, publications_mut, contacts_mut, studies_mut, remarks_mut, line_number_mut)
            (pattern_matching_result,) = (None,)
            if last_line is not None:
                if last_line == "ONTOLOGY SOURCE REFERENCE":
                    pattern_matching_result = 0

                elif last_line == "INVESTIGATION":
                    pattern_matching_result = 1

                elif last_line == "INVESTIGATION PUBLICATIONS":
                    pattern_matching_result = 2

                elif last_line == "INVESTIGATION CONTACTS":
                    pattern_matching_result = 3

                elif last_line == "STUDY":
                    pattern_matching_result = 4

                else: 
                    pattern_matching_result = 5


            else: 
                pattern_matching_result = 5

            if pattern_matching_result == 0:
                pattern_input: tuple[str | None, int, FSharpList[Remark], FSharpList[OntologySourceReference]] = from_rows(line_number + 1, en)
                last_line_mut = pattern_input[0]
                ontology_source_references_mut = pattern_input[3]
                investigation_info_mut = investigation_info
                publications_mut = publications
                contacts_mut = contacts
                studies_mut = studies
                remarks_mut = append(remarks, pattern_input[2])
                line_number_mut = pattern_input[1]
                continue

            elif pattern_matching_result == 1:
                pattern_input_1: tuple[str | None, int, FSharpList[Remark], ArcInvestigation_InvestigationInfo] = ArcInvestigation_InvestigationInfo_fromRows(line_number + 1, en)
                last_line_mut = pattern_input_1[0]
                ontology_source_references_mut = ontology_source_references
                investigation_info_mut = pattern_input_1[3]
                publications_mut = publications
                contacts_mut = contacts
                studies_mut = studies
                remarks_mut = append(remarks, pattern_input_1[2])
                line_number_mut = pattern_input_1[1]
                continue

            elif pattern_matching_result == 2:
                pattern_input_2: tuple[str | None, int, FSharpList[Remark], FSharpList[Publication]] = from_rows_1("Investigation Publication", line_number + 1, en)
                last_line_mut = pattern_input_2[0]
                ontology_source_references_mut = ontology_source_references
                investigation_info_mut = investigation_info
                publications_mut = pattern_input_2[3]
                contacts_mut = contacts
                studies_mut = studies
                remarks_mut = append(remarks, pattern_input_2[2])
                line_number_mut = pattern_input_2[1]
                continue

            elif pattern_matching_result == 3:
                pattern_input_3: tuple[str | None, int, FSharpList[Remark], FSharpList[Person]] = from_rows_2("Investigation Person", line_number + 1, en)
                last_line_mut = pattern_input_3[0]
                ontology_source_references_mut = ontology_source_references
                investigation_info_mut = investigation_info
                publications_mut = publications
                contacts_mut = pattern_input_3[3]
                studies_mut = studies
                remarks_mut = append(remarks, pattern_input_3[2])
                line_number_mut = pattern_input_3[1]
                continue

            elif pattern_matching_result == 4:
                pattern_input_4: tuple[str | None, int, FSharpList[Remark], tuple[ArcStudy, FSharpList[ArcAssay]] | None] = from_rows_3(line_number + 1, en)
                study: tuple[ArcStudy, FSharpList[ArcAssay]] | None = pattern_input_4[3]
                new_remarks_4: FSharpList[Remark] = pattern_input_4[2]
                line_number_6: int = pattern_input_4[1] or 0
                current_line_4: str | None = pattern_input_4[0]
                if study is not None:
                    last_line_mut = current_line_4
                    ontology_source_references_mut = ontology_source_references
                    investigation_info_mut = investigation_info
                    publications_mut = publications
                    contacts_mut = contacts
                    studies_mut = cons(value_1(study), studies)
                    remarks_mut = append(remarks, new_remarks_4)
                    line_number_mut = line_number_6
                    continue

                else: 
                    last_line_mut = current_line_4
                    ontology_source_references_mut = ontology_source_references
                    investigation_info_mut = investigation_info
                    publications_mut = publications
                    contacts_mut = contacts
                    studies_mut = studies
                    remarks_mut = append(remarks, new_remarks_4)
                    line_number_mut = line_number_6
                    continue


            elif pattern_matching_result == 5:
                if en.System_Collections_IEnumerator_MoveNext():
                    last_line_mut = SparseRowModule_tryGetValueAt(0, en.System_Collections_Generic_IEnumerator_1_get_Current())
                    ontology_source_references_mut = ontology_source_references
                    investigation_info_mut = investigation_info
                    publications_mut = publications
                    contacts_mut = contacts
                    studies_mut = studies
                    remarks_mut = remarks
                    line_number_mut = line_number
                    continue

                else: 
                    pattern_input_5: tuple[FSharpList[ArcStudy], FSharpList[ArcAssay]]
                    tupled_arg: tuple[FSharpList[ArcStudy], FSharpList[FSharpList[ArcAssay]]] = unzip(studies)
                    def projection(a_1: ArcAssay, last_line: Any=last_line, ontology_source_references: Any=ontology_source_references, investigation_info: Any=investigation_info, publications: Any=publications, contacts: Any=contacts, studies: Any=studies, remarks: Any=remarks, line_number: Any=line_number) -> str:
                        return a_1.Identifier

                    class ObjectExpr3653:
                        @property
                        def Equals(self) -> Callable[[str, str], bool]:
                            def _arrow3652(x: str, y: str) -> bool:
                                return x == y

                            return _arrow3652

                        @property
                        def GetHashCode(self) -> Callable[[str], int]:
                            return string_hash

                    pattern_input_5 = (reverse(tupled_arg[0]), List_distinctBy(projection, concat(tupled_arg[1]), ObjectExpr3653()))
                    return ArcInvestigation_fromParts(investigation_info, ontology_source_references, publications, contacts, pattern_input_5[0], pattern_input_5[1], remarks)


            break

    arc_investigation: ArcInvestigation
    ignore(en.System_Collections_IEnumerator_MoveNext())
    arc_investigation = loop(SparseRowModule_tryGetValueAt(0, en.System_Collections_Generic_IEnumerator_1_get_Current()), empty(), empty_investigation_info, empty(), empty(), empty(), empty(), 1)
    if arc_investigation.Identifier == "":
        raise Exception("Mandatory Investigation identifier is not present")

    return arc_investigation


def ArcInvestigation_toRows(investigation: ArcInvestigation) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    def _arrow3666(__unit: None=None, investigation: Any=investigation) -> FSharpList[IEnumerable_1[tuple[int, str]]]:
        remarks: FSharpList[Remark] = of_seq(investigation.Remarks)
        def _arrow3664(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
            def _arrow3663(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                def _arrow3662(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                    def _arrow3661(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                        def _arrow3660(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                            def _arrow3659(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                def _arrow3658(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                    def _arrow3657(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                        def _arrow3656(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                            def _arrow3655(study_identifier: str) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                                study: ArcStudy = default_arg(investigation.TryGetStudy(study_identifier), ArcStudy(study_identifier))
                                                def _arrow3654(__unit: None=None) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
                                                    return to_rows_3(study, None)

                                                return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["STUDY"]))), delay(_arrow3654))

                                            return collect(_arrow3655, investigation.RegisteredStudyIdentifiers)

                                        return append_1(to_rows_2("Investigation Person", of_seq(investigation.Contacts)), delay(_arrow3656))

                                    return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["INVESTIGATION CONTACTS"]))), delay(_arrow3657))

                                return append_1(to_rows_1("Investigation Publication", of_seq(investigation.Publications)), delay(_arrow3658))

                            return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["INVESTIGATION PUBLICATIONS"]))), delay(_arrow3659))

                        return append_1(ArcInvestigation_InvestigationInfo_toRows_Z720BD3FF(investigation), delay(_arrow3660))

                    return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["INVESTIGATION"]))), delay(_arrow3661))

                return append_1(to_rows(of_seq(investigation.OntologySourceReferences)), delay(_arrow3662))

            return append_1(singleton(SparseRowModule_fromValues(to_enumerable(["ONTOLOGY SOURCE REFERENCE"]))), delay(_arrow3663))

        rows: IEnumerable_1[IEnumerable_1[tuple[int, str]]] = delay(_arrow3664)
        try: 
            def mapping(remark: Remark) -> tuple[int, str]:
                return Remark.to_tuple(remark)

            class ObjectExpr3665:
                @property
                def Compare(self) -> Callable[[int, int], int]:
                    return compare_primitives

            rm: Any = of_list(map(mapping, remarks), ObjectExpr3665())
            def loop(i_mut: int, l_mut: FSharpList[IEnumerable_1[tuple[int, str]]], nl_mut: FSharpList[IEnumerable_1[tuple[int, str]]]) -> FSharpList[IEnumerable_1[tuple[int, str]]]:
                while True:
                    (i, l, nl) = (i_mut, l_mut, nl_mut)
                    match_value: str | None = try_find(i, rm)
                    if match_value is None:
                        if not is_empty_1(l):
                            i_mut = i + 1
                            l_mut = tail(l)
                            nl_mut = cons(head(l), nl)
                            continue

                        else: 
                            return nl


                    else: 
                        remark_1: str = match_value
                        i_mut = i + 1
                        l_mut = l
                        nl_mut = cons(SparseRowModule_fromValues(to_enumerable([Remark_wrapRemark(remark_1)])), nl)
                        continue

                    break

            return reverse(loop(1, of_seq(rows), empty()))

        except Exception as match_value_1:
            return to_list(rows)


    return _arrow3666()


def ArcInvestigation_toMetadataCollection(investigation: ArcInvestigation) -> IEnumerable_1[IEnumerable_1[str | None]]:
    def mapping(row: IEnumerable_1[tuple[int, str]], investigation: Any=investigation) -> IEnumerable_1[str | None]:
        return SparseRowModule_getAllValues(row)

    return map_1(mapping, ArcInvestigation_toRows(investigation))


def ArcInvestigation_fromMetadataCollection(collection: IEnumerable_1[IEnumerable_1[str | None]]) -> ArcInvestigation:
    def mapping(v: IEnumerable_1[str | None], collection: Any=collection) -> IEnumerable_1[tuple[int, str]]:
        return SparseRowModule_fromAllValues(v)

    return ArcInvestigation_fromRows(map_1(mapping, collection))


def ArcInvestigation_isMetadataSheetName(name: str) -> bool:
    if name == "isa_investigation":
        return True

    else: 
        return name == "Investigation"



def ArcInvestigation_isMetadataSheet(sheet: FsWorksheet) -> bool:
    return ArcInvestigation_isMetadataSheetName(sheet.Name)


def ArcInvestigation_tryGetMetadataSheet(doc: FsWorkbook) -> FsWorksheet | None:
    def predicate(sheet: FsWorksheet, doc: Any=doc) -> bool:
        return ArcInvestigation_isMetadataSheet(sheet)

    return try_find_1(predicate, doc.GetWorksheets())


def ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D(doc: FsWorkbook) -> ArcInvestigation:
    try: 
        def _arrow3667(__unit: None=None) -> Array[FsRow]:
            sheet_1: FsWorksheet
            match_value: FsWorksheet | None = ArcInvestigation_tryGetMetadataSheet(doc)
            if match_value is None:
                raise Exception("Could not find metadata sheet with sheetname \"isa_investigation\" or deprecated sheetname \"Investigation\"")

            else: 
                sheet_1 = match_value

            return FsWorksheet.get_rows(sheet_1)

        return ArcInvestigation_fromRows(map_1(SparseRowModule_fromFsRow, _arrow3667()))

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Could not read investigation from spreadsheet: %s"))(arg)



def ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(investigation: ArcInvestigation) -> FsWorkbook:
    try: 
        wb: FsWorkbook = FsWorkbook()
        sheet: FsWorksheet = FsWorksheet("isa_investigation")
        def action(row_i: int, r: IEnumerable_1[tuple[int, str]]) -> None:
            SparseRowModule_writeToSheet(row_i + 1, r, sheet)

        iterate_indexed(action, ArcInvestigation_toRows(investigation))
        wb.AddWorksheet(sheet)
        return wb

    except Exception as err:
        arg: str = str(err)
        return to_fail(printf("Could not write investigation to spreadsheet: %s"))(arg)



def ARCtrl_ArcInvestigation__ArcInvestigation_ToFsWorkbook(this: ArcInvestigation) -> FsWorkbook:
    return ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF(this)


__all__ = ["ArcInvestigation_InvestigationInfo_reflection", "ArcInvestigation_InvestigationInfo_create", "ArcInvestigation_InvestigationInfo_get_Labels", "ArcInvestigation_InvestigationInfo_FromSparseTable_3ECCA699", "ArcInvestigation_InvestigationInfo_ToSparseTable_Z720BD3FF", "ArcInvestigation_InvestigationInfo_fromRows", "ArcInvestigation_InvestigationInfo_toRows_Z720BD3FF", "ArcInvestigation_fromParts", "ArcInvestigation_fromRows", "ArcInvestigation_toRows", "ArcInvestigation_toMetadataCollection", "ArcInvestigation_fromMetadataCollection", "ArcInvestigation_isMetadataSheetName", "ArcInvestigation_isMetadataSheet", "ArcInvestigation_tryGetMetadataSheet", "ARCtrl_ArcInvestigation__ArcInvestigation_fromFsWorkbook_Static_32154C9D", "ARCtrl_ArcInvestigation__ArcInvestigation_toFsWorkbook_Static_Z720BD3FF", "ARCtrl_ArcInvestigation__ArcInvestigation_ToFsWorkbook"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...fable_library.list import (empty, FSharpList, map, append)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.string_ import (to_text, printf)
from .expression import (RequiredSource_1__ctor_2B595, RequiredSource_1, OptionalSource_1__ctor_2B595, OptionalSource_1, RequiredSource_1__get_Source, OptionalSource_1__get_Source)
from .types import (SheetEntity_1_some_2B595, SheetEntity_1, RowElement, Message__MapText_11D407F6, Message, SheetEntity_1__get_Messages)

def _expr312() -> TypeInfo:
    return class_type("FsSpreadsheet.DSL.RowBuilder", None, RowBuilder)


class RowBuilder:
    def __init__(self, __unit: None=None) -> None:
        pass


RowBuilder_reflection = _expr312

def RowBuilder__ctor(__unit: None=None) -> RowBuilder:
    return RowBuilder(__unit)


def RowBuilder_get_Empty(__unit: None=None) -> SheetEntity_1[FSharpList[RowElement]]:
    return SheetEntity_1_some_2B595(empty())


def RowBuilder__SignMessages_3F89FFFC(this: RowBuilder, messages: FSharpList[Message]) -> FSharpList[Message]:
    def mapping(m: Message, this: Any=this, messages: Any=messages) -> Message:
        def _arrow313(__unit: None=None, m: Any=m) -> Callable[[str], str]:
            clo: Callable[[str], str] = to_text(printf("In Row: %s"))
            return clo

        return Message__MapText_11D407F6(m, _arrow313())

    return map(mapping, messages)


def RowBuilder__Combine_19F30600(this: RowBuilder, wx1: SheetEntity_1[FSharpList[RowElement]], wx2: SheetEntity_1[FSharpList[RowElement]]) -> SheetEntity_1[FSharpList[RowElement]]:
    (pattern_matching_result, l1, l2, messages1, messages2, messages2_1, messages1_1, f1, messages1_2, messages2_2, f2, messages1_3, messages2_3, messages1_4, messages2_4) = (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)
    if wx1.tag == 2:
        if wx2.tag == 2:
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]

        else: 
            pattern_matching_result = 2
            messages1_1 = wx1.fields[0]


    elif wx1.tag == 1:
        if wx2.tag == 0:
            pattern_matching_result = 4
            f2 = wx2.fields[0]
            messages1_3 = wx1.fields[0]
            messages2_3 = wx2.fields[1]

        elif wx2.tag == 1:
            pattern_matching_result = 5
            messages1_4 = wx1.fields[0]
            messages2_4 = wx2.fields[0]

        else: 
            pattern_matching_result = 1
            messages2_1 = wx2.fields[0]


    elif wx2.tag == 2:
        pattern_matching_result = 1
        messages2_1 = wx2.fields[0]

    elif wx2.tag == 1:
        pattern_matching_result = 3
        f1 = wx1.fields[0]
        messages1_2 = wx1.fields[1]
        messages2_2 = wx2.fields[0]

    else: 
        pattern_matching_result = 0
        l1 = wx1.fields[0]
        l2 = wx2.fields[0]
        messages1 = wx1.fields[1]
        messages2 = wx2.fields[1]

    if pattern_matching_result == 0:
        return SheetEntity_1(0, append(l1, l2), append(messages1, messages2))

    elif pattern_matching_result == 1:
        return SheetEntity_1(2, append(SheetEntity_1__get_Messages(wx1), messages2_1))

    elif pattern_matching_result == 2:
        return SheetEntity_1(2, append(messages1_1, SheetEntity_1__get_Messages(wx2)))

    elif pattern_matching_result == 3:
        return SheetEntity_1(0, f1, append(messages1_2, messages2_2))

    elif pattern_matching_result == 4:
        return SheetEntity_1(0, f2, append(messages1_3, messages2_3))

    elif pattern_matching_result == 5:
        return SheetEntity_1(1, append(messages1_4, messages2_4))



def RowBuilder__Combine_ZC028CBD(this: RowBuilder, wx1: RequiredSource_1[None], wx2: SheetEntity_1[FSharpList[RowElement]]) -> RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return RequiredSource_1__ctor_2B595(wx2)


def RowBuilder__Combine_7DC18FE3(this: RowBuilder, wx1: SheetEntity_1[FSharpList[RowElement]], wx2: RequiredSource_1[None]) -> RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return RequiredSource_1__ctor_2B595(wx1)


def RowBuilder__Combine_Z1FF3C0DC(this: RowBuilder, wx1: OptionalSource_1[None], wx2: SheetEntity_1[FSharpList[RowElement]]) -> OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return OptionalSource_1__ctor_2B595(wx2)


def RowBuilder__Combine_Z47B5345C(this: RowBuilder, wx1: SheetEntity_1[FSharpList[RowElement]], wx2: OptionalSource_1[None]) -> OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return OptionalSource_1__ctor_2B595(wx1)


def RowBuilder__Combine_385AF66F(this: RowBuilder, wx1: RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]], wx2: SheetEntity_1[FSharpList[RowElement]]) -> RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return RequiredSource_1__ctor_2B595(RowBuilder__Combine_19F30600(this, RequiredSource_1__get_Source(wx1), wx2))


def RowBuilder__Combine_16FE340F(this: RowBuilder, wx1: SheetEntity_1[FSharpList[RowElement]], wx2: RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]]) -> RequiredSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return RequiredSource_1__ctor_2B595(RowBuilder__Combine_19F30600(this, wx1, RequiredSource_1__get_Source(wx2)))


def RowBuilder__Combine_Z4C077C38(this: RowBuilder, wx1: OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]], wx2: SheetEntity_1[FSharpList[RowElement]]) -> OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return OptionalSource_1__ctor_2B595(RowBuilder__Combine_19F30600(this, OptionalSource_1__get_Source(wx1), wx2))


def RowBuilder__Combine_Z23359F38(this: RowBuilder, wx1: SheetEntity_1[FSharpList[RowElement]], wx2: OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]]) -> OptionalSource_1[SheetEntity_1[FSharpList[RowElement]]]:
    return OptionalSource_1__ctor_2B595(RowBuilder__Combine_19F30600(this, wx1, OptionalSource_1__get_Source(wx2)))


__all__ = ["RowBuilder_reflection", "RowBuilder_get_Empty", "RowBuilder__SignMessages_3F89FFFC", "RowBuilder__Combine_19F30600", "RowBuilder__Combine_ZC028CBD", "RowBuilder__Combine_7DC18FE3", "RowBuilder__Combine_Z1FF3C0DC", "RowBuilder__Combine_Z47B5345C", "RowBuilder__Combine_385AF66F", "RowBuilder__Combine_16FE340F", "RowBuilder__Combine_Z4C077C38", "RowBuilder__Combine_Z23359F38"]




from __future__ import annotations
from typing import (Any, TypeVar)
from ...arctrl_core.data import Data
from ...arctrl_core.data_context import (DataContext__get_Explication, DataContext__get_Unit, DataContext__get_ObjectType, DataContext__get_Description, DataContext__get_GeneratedBy, DataContext__get_Label, DataContext, DataContext__ctor_Z780A8A2A)
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (choose, of_array, FSharpList)
from ...fable_library.option import map
from ...fable_library.seq import map as map_1
from ...fable_library.util import IEnumerable_1
from ...thoth_json_core.decode import (object, IRequiredGetter, IOptionalGetter, string, IGetters)
from ...thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from ..data import (encoder as encoder_1, decoder as decoder_1)
from ..encode import try_include
from ..ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder)

__A_ = TypeVar("__A_")

def encoder(dc: DataContext) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], dc: Any=dc) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2817(oa: OntologyAnnotation, dc: Any=dc) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2821(oa_1: OntologyAnnotation, dc: Any=dc) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2822(oa_2: OntologyAnnotation, dc: Any=dc) -> IEncodable:
        return OntologyAnnotation_encoder(oa_2)

    def _arrow2828(value: str, dc: Any=dc) -> IEncodable:
        class ObjectExpr2827(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2827()

    def _arrow2830(value_2: str, dc: Any=dc) -> IEncodable:
        class ObjectExpr2829(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_2)

        return ObjectExpr2829()

    def _arrow2835(value_4: str, dc: Any=dc) -> IEncodable:
        class ObjectExpr2834(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_4)

        return ObjectExpr2834()

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("data", encoder_1(dc)), try_include("explication", _arrow2817, DataContext__get_Explication(dc)), try_include("unit", _arrow2821, DataContext__get_Unit(dc)), try_include("objectType", _arrow2822, DataContext__get_ObjectType(dc)), try_include("description", _arrow2828, DataContext__get_Description(dc)), try_include("generatedBy", _arrow2830, DataContext__get_GeneratedBy(dc)), try_include("label", _arrow2835, DataContext__get_Label(dc))]))
    class ObjectExpr2837(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], dc: Any=dc) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr2837()


def _arrow2855(get: IGetters) -> DataContext:
    data: Data
    object_arg: IRequiredGetter = get.Required
    data = object_arg.Field("data", decoder_1)
    def _arrow2846(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("explication", OntologyAnnotation_decoder)

    def _arrow2848(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("unit", OntologyAnnotation_decoder)

    def _arrow2849(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("objectType", OntologyAnnotation_decoder)

    def _arrow2852(__unit: None=None) -> str | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("label", string)

    def _arrow2853(__unit: None=None) -> str | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("description", string)

    def _arrow2854(__unit: None=None) -> str | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("generatedBy", string)

    return DataContext__ctor_Z780A8A2A(data.ID, data.Name, data.DataType, data.Format, data.SelectorFormat, _arrow2846(), _arrow2848(), _arrow2849(), _arrow2852(), _arrow2853(), _arrow2854(), data.Comments)


decoder: Decoder_1[DataContext] = object(_arrow2855)

__all__ = ["encoder", "decoder"]

